import base64
import datetime
import hashlib
import hmac
import re
import time
from typing import Any, Dict, Optional
from urllib.parse import urlencode

import hummingbot.connector.derivative.okx_perpetual.okx_perpetual_constants as CONSTANTS
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest, WSRequest


class OkxPerpetualAuth(AuthBase):
    """
    Auth class required by OKX Perpetual API.

    All private REST requests must contain the following headers:

        - "OK-ACCESS-KEY" -> The API Key as a String.
        - "OK-ACCESS-SIGN" -> The Base64-encoded signature.
        - "OK-ACCESS-TIMESTAMP" -> The UTC timestamp of your request .e.g : 2020-12-08T09:08:57.715Z
        - "OK-ACCESS-PASSPHRASE" -> The passphrase you specified when creating the APIKey.

        Request bodies should have content type application/json and be in valid JSON format.
    """

    def __init__(self, api_key: str, api_secret: str, passphrase: str, time_provider: TimeSynchronizer):
        self._api_key: str = api_key
        self._api_secret: str = api_secret
        self._passphrase: str = passphrase
        self.time_provider: TimeSynchronizer = time_provider

    def _generate_signature(self, timestamp: str, method: str, path_url: str, body: Optional[str] = None) -> str:
        unsigned_signature = timestamp + method + path_url
        if body is not None:
            unsigned_signature += body

        signature = base64.b64encode(
            hmac.new(
                self._api_secret.encode("utf-8"),
                unsigned_signature.encode("utf-8"),
                hashlib.sha256).digest()).decode()
        return signature

    def authentication_headers(self, request: RESTRequest) -> Dict[str, Any]:
        timestamp = datetime.datetime.fromtimestamp(self.time_provider.time(), datetime.UTC).isoformat(timespec="milliseconds")
        timestamp = timestamp.replace("+00:00", "Z")

        path_url = f"/api{request.url.split('/api')[-1]}"
        if request.params:
            query_string_components = urlencode(request.params)
            query_string_components_with_comma = query_string_components.replace("%2C", ",")
            path_url = f"{path_url}?{query_string_components_with_comma}"

        header = {
            "OK-ACCESS-KEY": self._api_key,
            "OK-ACCESS-SIGN": self._generate_signature(timestamp, request.method.value.upper(), path_url, request.data),
            "OK-ACCESS-TIMESTAMP": timestamp,
            "OK-ACCESS-PASSPHRASE": self._passphrase,
        }

        return header

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        """
        All private REST requests must contain the following headers:

            - OK-ACCESS-KEY The API Key as a String.
            - OK-ACCESS-SIGN The Base64-encoded signature
            - OK-ACCESS-TIMESTAMP The UTC timestamp of your request .e.g : 2020-12-08T09:08:57.715Z
            - OK-ACCESS-PASSPHRASE The passphrase you specified when creating the APIKey.

        Request bodies should have content type application/json and be in valid JSON format.
        """

        headers = {}
        if request.headers is not None:
            headers.update(request.headers)
        headers.update(self.authentication_headers(request=request))
        request.headers = headers

        return request

    @staticmethod
    def get_path_from_url(url: str) -> str:
        """
        The requestPath is the path of requesting an endpoint.

            - Example: /api/v5/account/balance

        """
        pattern = re.compile(r'https://www.okx.com')
        return re.sub(pattern, '', url)

    def get_ws_auth_args(self) -> Dict[str, str]:
        """
            - api_key: Unique identification for invoking API. Requires user to apply one manually.
            - passphrase: API Key password
            - timestamp: the Unix Epoch time, the unit is seconds
            - sign: signature string, the signature algorithm is as follows:

        First concatenate timestamp, method, requestPath, strings, then use HMAC SHA256 method to encrypt
        the concatenated string with SecretKey, and then perform Base64 encoding.
        """
        timestamp = int(time.time())
        _access_sign = self.generate_ws_signature_from_payload(timestamp=timestamp,
                                                               method=RESTMethod.GET,
                                                               request_path=CONSTANTS.REST_WS_LOGIN_PATH["ENDPOINT"])
        return [
            {
                "apiKey": self._api_key,
                "passphrase": self._passphrase,
                "timestamp": timestamp,
                "sign": _access_sign
            }
        ]

    def generate_ws_signature_from_payload(self, timestamp: int, method: RESTMethod, request_path: str) -> str:
        message = str(timestamp) + str.upper(method.value) + request_path
        mac = hmac.new(bytes(self._api_secret, encoding='utf8'), bytes(message, encoding='utf-8'), digestmod='sha256')
        d = mac.digest()
        return str(base64.b64encode(d), encoding='utf-8')

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        return request  # pass-through

    @staticmethod
    def _get_timestamp() -> str:
        return datetime.datetime.now(datetime.UTC).isoformat(timespec='milliseconds')



import asyncio
import logging
import time
from collections import deque
from typing import Deque, Dict, List, Optional, Union

from sqlalchemy.orm import Session

from hummingbot.client.command import __all__ as commands
from hummingbot.client.config.client_config_map import ClientConfigMap
from hummingbot.client.config.config_helpers import (
    ClientConfigAdapter,
    get_strategy_config_map,
    load_client_config_map_from_file,
    load_ssl_config_map_from_file,
    save_to_yml,
)
from hummingbot.client.config.gateway_ssl_config_map import SSLConfigMap
from hummingbot.client.config.strategy_config_data_types import BaseStrategyConfigMap
from hummingbot.client.settings import CLIENT_CONFIG_PATH
from hummingbot.client.tab import __all__ as tab_classes
from hummingbot.client.tab.data_types import CommandTab
from hummingbot.client.ui.completer import load_completer
from hummingbot.client.ui.hummingbot_cli import HummingbotCLI
from hummingbot.client.ui.keybindings import load_key_bindings
from hummingbot.client.ui.parser import ThrowingArgumentParser, load_parser
from hummingbot.connector.exchange_base import ExchangeBase
from hummingbot.core.trading_core import TradingCore
from hummingbot.core.utils.trading_pair_fetcher import TradingPairFetcher
from hummingbot.exceptions import ArgumentParserError
from hummingbot.logger import HummingbotLogger
from hummingbot.logger.application_warning import ApplicationWarning
from hummingbot.model.trade_fill import TradeFill
from hummingbot.remote_iface.mqtt import MQTTGateway

s_logger = None


class HummingbotApplication(*commands):
    KILL_TIMEOUT = 20.0
    APP_WARNING_EXPIRY_DURATION = 3600.0
    APP_WARNING_STATUS_LIMIT = 6

    _main_app: Optional["HummingbotApplication"] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        global s_logger
        if s_logger is None:
            s_logger = logging.getLogger(__name__)
        return s_logger

    @classmethod
    def main_application(cls, client_config_map: Optional[ClientConfigAdapter] = None, headless_mode: bool = False) -> "HummingbotApplication":
        if cls._main_app is None:
            cls._main_app = HummingbotApplication(client_config_map=client_config_map, headless_mode=headless_mode)
        return cls._main_app

    def __init__(self, client_config_map: Optional[ClientConfigAdapter] = None, headless_mode: bool = False):
        self.client_config_map: Union[ClientConfigMap, ClientConfigAdapter] = (  # type-hint enables IDE auto-complete
            client_config_map or load_client_config_map_from_file()
        )
        self.headless_mode = headless_mode
        self.ssl_config_map: SSLConfigMap = (  # type-hint enables IDE auto-complete
            load_ssl_config_map_from_file()
        )
        self.ev_loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        # Initialize core trading functionality
        self.trading_core = TradingCore(self.client_config_map)

        # Application-specific properties
        self.init_time: float = time.time()
        self.placeholder_mode = False
        self._app_warnings: Deque[ApplicationWarning] = deque()

        # MQTT management
        self._mqtt: Optional[MQTTGateway] = None

        # Script configuration support
        self.script_config: Optional[str] = None

        # Initialize UI components only if not in headless mode
        if not headless_mode:
            self._init_ui_components()
            TradingPairFetcher.get_instance(self.client_config_map)
        else:
            # In headless mode, we don't initialize UI components
            self.app = None
            self.parser = None

        # MQTT Bridge (always available in both modes)
        if self.client_config_map.mqtt_bridge.mqtt_autostart:
            self.mqtt_start()

    def _init_ui_components(self):
        """Initialize UI components (CLI, parser, etc.) for non-headless mode."""
        command_tabs = self.init_command_tabs()
        self.parser: ThrowingArgumentParser = load_parser(self, command_tabs)
        self.app = HummingbotCLI(
            self.client_config_map,
            input_handler=self._handle_command,
            bindings=load_key_bindings(self),
            completer=load_completer(self),
            command_tabs=command_tabs
        )

    @property
    def instance_id(self) -> str:
        return self.client_config_map.instance_id

    @property
    def fetch_pairs_from_all_exchanges(self) -> bool:
        return self.client_config_map.fetch_pairs_from_all_exchanges

    @property
    def gateway_config_keys(self) -> List[str]:
        return self.trading_core.gateway_monitor.gateway_config_keys

    @property
    def strategy_file_name(self) -> str:
        return self.trading_core.strategy_file_name

    @strategy_file_name.setter
    def strategy_file_name(self, value: Optional[str]):
        self.trading_core.strategy_file_name = value

    @property
    def strategy_name(self) -> str:
        return self.trading_core.strategy_name

    @strategy_name.setter
    def strategy_name(self, value: Optional[str]):
        self.trading_core.strategy_name = value

    @property
    def markets(self) -> Dict[str, ExchangeBase]:
        return self.trading_core.markets

    @property
    def notifiers(self):
        return self.trading_core.notifiers

    @property
    def strategy_config_map(self):
        if self.trading_core.strategy_config_map is not None:
            return self.trading_core.strategy_config_map
        if self.trading_core.strategy_name is not None:
            return get_strategy_config_map(self.trading_core.strategy_name)
        return None

    @strategy_config_map.setter
    def strategy_config_map(self, config_map: BaseStrategyConfigMap):
        self.trading_core.strategy_config_map = config_map

    def notify(self, msg: str):
        # In headless mode, just log to console and notifiers
        if self.headless_mode:
            self.logger().info(msg)
        else:
            self.app.log(msg)
        for notifier in self.trading_core.notifiers:
            notifier.add_message_to_queue(msg)

    def _handle_command(self, raw_command: str):
        # unset to_stop_config flag it triggered before loading any command (UI mode only)
        if not self.headless_mode and hasattr(self, 'app') and self.app.to_stop_config:
            self.app.to_stop_config = False

        raw_command = raw_command.strip()
        # NOTE: Only done for config command
        if raw_command.startswith("config"):
            command_split = raw_command.split(maxsplit=2)
        else:
            command_split = raw_command.split()
        try:
            if self.placeholder_mode:
                pass
            elif len(command_split) == 0:
                pass
            else:
                # Check if help is requested, if yes, print & terminate
                if len(command_split) > 1 and any(arg in ["-h", "--help"] for arg in command_split[1:]):
                    self.help(raw_command)
                    return

                # regular command
                if self.headless_mode and not hasattr(self, 'parser'):
                    self.notify("Command parsing not available in headless mode")
                    return

                args = self.parser.parse_args(args=command_split)
                kwargs = vars(args)

                if not hasattr(args, "func"):
                    if not self.headless_mode:
                        self.app.handle_tab_command(self, command_split[0], kwargs)
                    else:
                        self.notify(f"Tab command '{command_split[0]}' not available in headless mode")
                else:
                    f = args.func
                    del kwargs["func"]
                    f(**kwargs)
        except ArgumentParserError as e:
            if not self.be_silly(raw_command):
                self.notify(str(e))
        except NotImplementedError:
            self.notify("Command not yet implemented. This feature is currently under development.")
        except Exception as e:
            self.logger().error(e, exc_info=True)

    async def run(self):
        """Run the application - either UI mode or headless mode."""
        if self.headless_mode:
            # Start MQTT market events forwarding if MQTT is available
            if self._mqtt is not None:
                self._mqtt.start_market_events_fw()
            await self.run_headless()
        else:
            await self.app.run()

    async def run_headless(self):
        """Run in headless mode - just keep alive for MQTT/strategy execution."""
        try:
            self.logger().info("Starting Hummingbot in headless mode...")

            # Validate MQTT is enabled for headless mode
            if not self.client_config_map.mqtt_bridge.mqtt_autostart:
                error_msg = (
                    "ERROR: MQTT must be enabled for headless mode!\n"
                    "Without MQTT, there would be no way to control the bot.\n"
                    "Please enable MQTT by setting 'mqtt_autostart: true' in your config file.\n"
                    "You can also start it manually with 'mqtt start' before switching to headless mode."
                )
                self.logger().error(error_msg)
                raise RuntimeError("MQTT is required for headless mode")

            self.logger().info("MQTT enabled - waiting for MQTT commands...")
            self.logger().info("Bot is ready to receive commands via MQTT")

            # Keep running until shutdown
            while True:
                await asyncio.sleep(1)

        except KeyboardInterrupt:
            self.logger().info("Shutdown requested...")
        except Exception as e:
            self.logger().error(f"Error in headless mode: {e}")
            raise
        finally:
            await self.trading_core.shutdown()

    def add_application_warning(self, app_warning: ApplicationWarning):
        self._expire_old_application_warnings()
        self._app_warnings.append(app_warning)

    def clear_application_warning(self):
        self._app_warnings.clear()

    def _initialize_notifiers(self):
        """Initialize notifiers by delegating to TradingCore."""
        for notifier in self.trading_core.notifiers:
            notifier.start()

    def init_command_tabs(self) -> Dict[str, CommandTab]:
        """
        Initiates and returns a CommandTab dictionary with mostly defaults and None values, These values will be
        populated later on by HummingbotCLI
        """
        command_tabs: Dict[str, CommandTab] = {}
        for tab_class in tab_classes:
            name = tab_class.get_command_name()
            command_tabs[name] = CommandTab(name, None, None, None, tab_class)
        return command_tabs

    def _get_trades_from_session(self,
                                 start_timestamp: int,
                                 session: Session,
                                 number_of_rows: Optional[int] = None,
                                 config_file_path: str = None) -> List[TradeFill]:
        return self.trading_core._get_trades_from_session(start_timestamp, session, number_of_rows, config_file_path)

    def save_client_config(self):
        save_to_yml(CLIENT_CONFIG_PATH, self.client_config_map)



import ecdsa
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from xrpl.constants import CryptoAlgorithm
from xrpl.wallet import Wallet

from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTRequest, WSRequest


class XRPLAuth(AuthBase):
    def __init__(self, xrpl_secret_key: str):
        try:
            if len(xrpl_secret_key) == 0:
                self._wallet = Wallet.create()
            else:
                # Check if it's a raw private key (hex format)
                if xrpl_secret_key.startswith("ED"):
                    # Raw private key for Ed25519
                    private_key = xrpl_secret_key
                    # Convert private key to bytes
                    private_key_bytes = bytes.fromhex(private_key[2:])  # Remove ED prefix
                    # Derive public key using Ed25519
                    private_key_obj = Ed25519PrivateKey.from_private_bytes(private_key_bytes)
                    public_key_bytes = private_key_obj.public_key().public_bytes_raw()
                    public_key = "ED" + public_key_bytes.hex()
                    # Create wallet with derived keys
                    self._wallet = Wallet(public_key=public_key, private_key=private_key)
                elif xrpl_secret_key.startswith("00"):
                    # Raw private key for SECP256K1
                    private_key = xrpl_secret_key
                    # Convert private key to bytes
                    private_key_bytes = bytes.fromhex(private_key[2:])  # Remove 00 prefix
                    # Create private key object and derive public key using SECP256K1
                    private_key_obj = ecdsa.SigningKey.from_string(private_key_bytes, curve=ecdsa.SECP256k1)
                    verifying_key = private_key_obj.get_verifying_key()
                    if verifying_key is None:
                        raise ValueError("Failed to derive public key from SECP256K1 private key")
                    public_key_bytes = verifying_key.to_string()
                    public_key = "00" + public_key_bytes.hex().upper()
                    # Create wallet with derived keys
                    self._wallet = Wallet(public_key=public_key, private_key=private_key)
                elif xrpl_secret_key.startswith("s"):
                    # Seed format
                    self._wallet = Wallet.from_seed(xrpl_secret_key, algorithm=self.get_algorithm(key=xrpl_secret_key))
                else:
                    raise ValueError(
                        "Invalid XRPL secret key format. Must be either a seed (starting with 's'), or a raw private key (starting with 'ED' or '00')"
                    )
        except Exception as e:
            raise ValueError(f"Invalid XRPL secret key: {e}")

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        return request

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        return request

    def get_wallet(self) -> Wallet:
        return self._wallet

    def get_account(self) -> str:
        return self._wallet.classic_address

    def get_algorithm(self, key: str) -> CryptoAlgorithm:
        if key.startswith("sEd"):
            return CryptoAlgorithm.ED25519
        elif key.startswith("s"):
            return CryptoAlgorithm.SECP256K1
        else:
            raise ValueError("Invalid key format")



import sys

from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.common import OrderType, PositionMode
from hummingbot.core.data_type.in_flight_order import OrderState

EXCHANGE_NAME = "kucoin_perpetual"

DEFAULT_DOMAIN = "kucoin_perpetual_main"

DEFAULT_TIME_IN_FORCE = "GTC"

REST_URLS = {"kucoin_perpetual_main": "https://api-futures.kucoin.com/"}
WSS_PUBLIC_URLS = {"kucoin_perpetual_main": "wss://stream.kucoin.com/realtime_public"}
WSS_PRIVATE_URLS = {"kucoin_perpetual_main": "wss://stream.kucoin.com/realtime_private"}
REST_API_VERSION = "api/v1"

HB_PARTNER_ID = "Hummingbot"
HB_PARTNER_KEY = "8fb50686-81a8-408a-901c-07c5ac5bd758"

MAX_ID_LEN = 40
SECONDS_TO_WAIT_TO_RECEIVE_MESSAGE = 30

ONE_HOUR = 3600

WS_HEARTBEAT_TIME_INTERVAL = 30

ORDER_TYPE_MAP = {
    OrderType.LIMIT: "limit",
    OrderType.LIMIT_MAKER: "limit",
    OrderType.MARKET: "market",
}

POSITION_MODE_API_ONEWAY = "MergedSingle"
POSITION_MODE_API_HEDGE = "BothSide"
POSITION_MODE_MAP = {
    PositionMode.ONEWAY: POSITION_MODE_API_ONEWAY,
    PositionMode.HEDGE: POSITION_MODE_API_HEDGE,
}

# REST API Public Endpoints
QUERY_SYMBOL_ENDPOINT = f"{REST_API_VERSION}/contracts/active"
LATEST_SYMBOL_INFORMATION_ENDPOINT = f"{REST_API_VERSION}/ticker?symbol={{symbol}}"
ORDER_BOOK_ENDPOINT = f"{REST_API_VERSION}/level2/depth100?symbol={{symbol}}"
SERVER_TIME_PATH_URL = f"{REST_API_VERSION}/timestamp"
GET_LAST_FUNDING_RATE_PATH_URL = f"{REST_API_VERSION}/funding-rate/{{symbol}}/current"
GET_FUNDING_HISTORY_PATH_URL = f"{REST_API_VERSION}/funding-history?symbol={{symbol}}"
GET_CONTRACT_INFO_PATH_URL = f"{REST_API_VERSION}/contracts/{{symbol}}"

# REST API Private Endpoints
CREATE_ORDER_PATH_URL = f"{REST_API_VERSION}/orders"
CANCEL_ORDER_PATH_URL = f"{REST_API_VERSION}/orders/{{orderid}}"
QUERY_ORDER_BY_EXCHANGE_ORDER_ID_PATH_URL = f"{REST_API_VERSION}/orders/{{orderid}}"
QUERY_ORDER_BY_CLIENT_ORDER_ID_PATH_URL = f"{REST_API_VERSION}/orders/byClientOid?clientOid={{clientorderid}}"
GET_RISK_LIMIT_LEVEL_PATH_URL = f"{REST_API_VERSION}/contracts/risk-limit/{{symbol}}"
SET_LEVERAGE_PATH_URL = f"{REST_API_VERSION}/position/risk-limit-level/change"
GET_RECENT_FILLS_INFO_PATH_URL = f"{REST_API_VERSION}/recentFills"
GET_FILL_INFO_PATH_URL = f"{REST_API_VERSION}/fills?orderId={{orderid}}"
GET_WALLET_BALANCE_PATH_URL = f"{REST_API_VERSION}/account-overview?currency={{currency}}"
GET_POSITIONS_PATH_URL = f"{REST_API_VERSION}/positions"
QUERY_ACTIVE_ORDER_PATH_URL = f"{REST_API_VERSION}/orders?status=active"
QUERY_ALL_ORDER_PATH_URL = f"{REST_API_VERSION}/orders"

# Websocket
PUBLIC_WS_DATA_PATH_URL = f"{REST_API_VERSION}/bullet-public"
PRIVATE_WS_DATA_PATH_URL = f"{REST_API_VERSION}/bullet-private"

WS_PING_REQUEST = "ping"
WS_ORDER_BOOK_EVENTS_TOPIC = "/contractMarket/level2"
WS_TRADES_TOPIC = "/contractMarket/tradeOrders"
WS_INSTRUMENTS_INFO_TOPIC = "/contract/instrument"
WS_TICKER_INFO_TOPIC = "/contractMarket/ticker"
WS_WALLET_INFO_TOPIC = "/contractAccount/wallet"
WS_EXECUTION_DATA_TOPIC = "/contractMarket/execution"
WS_POSITION_CHANGE_TOPIC = "/contract/position"
WS_AUTHENTICATE_USER_ENDPOINT_NAME = "auth"
WS_SUBSCRIPTION_POSITIONS_ENDPOINT_NAME = "position.change"
WS_SUBSCRIPTION_ORDERS_ENDPOINT_NAME = "orderChange"
WS_SUBSCRIPTION_EXECUTIONS_ENDPOINT_NAME = "match"
WS_SUBSCRIPTION_WELCOME_MESSAGE = "welcome"
WS_ADJUST_RISK_LIMIT_MESSAGE = "position.adjustRiskLimit"

WS_SUBSCRIPTION_WALLET_ENDPOINT_NAME = "availableBalance.change"

WS_CONNECTION_LIMIT_ID = "WSConnection"
WS_CONNECTION_LIMIT = 30
WS_CONNECTION_TIME_INTERVAL = 20
WS_REQUEST_LIMIT_ID = "WSRequest"

# Order Statuses
ORDER_STATE = {
    "open": OrderState.OPEN,
    "done": OrderState.FILLED,
    "cancelExist": OrderState.CANCELED,
}

# Request error codes
RET_CODE_OK = "200000"
RET_CODE_PARAMS_ERROR = "100001"
RET_CODE_API_KEY_INVALID = "400001"
RET_CODE_ORDER_NOT_EXISTS = "20001"
RET_CODE_MODE_POSITION_NOT_EMPTY = "30082"
RET_CODE_MODE_NOT_MODIFIED = "300010"
RET_CODE_LEVERAGE_NOT_MODIFIED = "300016"
RET_CODE_POSITION_ZERO = "300009"
RET_CODE_AUTH_TIMESTAMP_ERROR = "400002"

NO_LIMIT = sys.maxsize
RATE_LIMITS = [
    RateLimit(WS_CONNECTION_LIMIT_ID, limit=WS_CONNECTION_LIMIT, time_interval=WS_CONNECTION_TIME_INTERVAL),
    RateLimit(WS_REQUEST_LIMIT_ID, limit=100, time_interval=10),
    RateLimit(limit_id=PUBLIC_WS_DATA_PATH_URL, limit=100, time_interval=10),
    RateLimit(limit_id=PRIVATE_WS_DATA_PATH_URL, limit=100, time_interval=10),
    RateLimit(limit_id=LATEST_SYMBOL_INFORMATION_ENDPOINT, limit=NO_LIMIT, time_interval=1),
    RateLimit(limit_id=QUERY_SYMBOL_ENDPOINT, limit=NO_LIMIT, time_interval=1),
    RateLimit(limit_id=ORDER_BOOK_ENDPOINT, limit=30, time_interval=3),
    RateLimit(limit_id=SERVER_TIME_PATH_URL, limit=NO_LIMIT, time_interval=1),
    RateLimit(limit_id=SET_LEVERAGE_PATH_URL, limit=NO_LIMIT, time_interval=1),
    RateLimit(limit_id=GET_LAST_FUNDING_RATE_PATH_URL, limit=9, time_interval=3),
    RateLimit(limit_id=GET_POSITIONS_PATH_URL, limit=9, time_interval=3),
    RateLimit(limit_id=QUERY_ORDER_BY_EXCHANGE_ORDER_ID_PATH_URL, limit=30, time_interval=3),
    RateLimit(limit_id=QUERY_ORDER_BY_CLIENT_ORDER_ID_PATH_URL, limit=30, time_interval=3),
    RateLimit(limit_id=QUERY_ACTIVE_ORDER_PATH_URL, limit=30, time_interval=3),
    RateLimit(limit_id=GET_WALLET_BALANCE_PATH_URL, limit=30, time_interval=3),
    RateLimit(limit_id=GET_CONTRACT_INFO_PATH_URL, limit=NO_LIMIT, time_interval=1),
    RateLimit(limit_id=CREATE_ORDER_PATH_URL, limit=25, time_interval=3),
    RateLimit(limit_id=CANCEL_ORDER_PATH_URL, limit=35, time_interval=3),
    RateLimit(limit_id=GET_FILL_INFO_PATH_URL, limit=9, time_interval=3),
    RateLimit(limit_id=GET_RECENT_FILLS_INFO_PATH_URL, limit=9, time_interval=3),
    RateLimit(limit_id=GET_FUNDING_HISTORY_PATH_URL, limit=9, time_interval=3),
    RateLimit(limit_id=GET_RISK_LIMIT_LEVEL_PATH_URL, limit=9, time_interval=3),
]



from bidict import bidict

from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit

REST_URL = "https://www.okx.com"
WSS_URL = "wss://ws.okx.com:8443/ws/v5/business"

# GET / Candlesticks history (https://www.okx.com/docs-v5/en/?shell#order-book-trading-market-data-get-candlesticks-history)
# Retrieve history candlestick charts from recent years(It is last 3 months supported for 1s candlestick).
# Rate Limit: 20 requests per 2 seconds
# Rate limit rule: IP
CANDLES_ENDPOINT = "/api/v5/market/history-candles"
INTERVALS = bidict({
    "1s": "1s",
    "1m": "1m",
    "3m": "3m",
    "5m": "5m",
    "15m": "15m",
    "30m": "30m",
    "1h": "1H",
    "2h": "2H",
    "4h": "4H",
    "6h": "6Hutc",
    "8h": "8Hutc",
    "12h": "12Hutc",
    "1d": "1Dutc",
    "3d": "3Dutc",
    "1w": "1Wutc",
    "1M": "1Mutc",
    "3M": "3Mutc"
})
MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST = 100

# Get system time (https://www.okx.com/docs-v5/en/?shell#public-data-rest-api-get-system-time)
# Retrieve API server time.
# Rate Limit: 10 requests per 2 seconds
# Rate limit rule: IP
HEALTH_CHECK_ENDPOINT = "/api/v5/public/time"

RATE_LIMITS = [
    RateLimit(CANDLES_ENDPOINT, limit=20, time_interval=2, linked_limits=[LinkedLimitWeightPair("raw", 1)]),
    RateLimit(HEALTH_CHECK_ENDPOINT, limit=10, time_interval=2, linked_limits=[LinkedLimitWeightPair("raw", 1)])]



import asyncio
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.exchange.kucoin import kucoin_constants as CONSTANTS, kucoin_web_utils as web_utils
from hummingbot.connector.exchange.kucoin.kucoin_auth import KucoinAuth
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.kucoin.kucoin_exchange import KucoinExchange


class KucoinAPIUserStreamDataSource(UserStreamTrackerDataSource):

    _logger: Optional[HummingbotLogger] = None

    def __init__(self,
                 auth: KucoinAuth,
                 trading_pairs: List[str],
                 connector: 'KucoinExchange',
                 api_factory: WebAssistantsFactory,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN):
        super().__init__()
        self._domain = domain
        self._api_factory = api_factory
        self._last_ws_message_sent_timestamp = 0
        self._ping_interval = 0

    async def _connected_websocket_assistant(self) -> WSAssistant:
        rest_assistant = await self._api_factory.get_rest_assistant()
        connection_info = await rest_assistant.execute_request(
            url=web_utils.private_rest_url(path_url=CONSTANTS.PRIVATE_WS_DATA_PATH_URL, domain=self._domain),
            method=RESTMethod.POST,
            throttler_limit_id=CONSTANTS.PRIVATE_WS_DATA_PATH_URL,
            is_auth_required=True,
        )

        ws_url = connection_info["data"]["instanceServers"][0]["endpoint"]
        self._ping_interval = int(connection_info["data"]["instanceServers"][0]["pingInterval"]) * 0.8 * 1e-3
        token = connection_info["data"]["token"]

        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=f"{ws_url}?token={token}", message_timeout=self._ping_interval)
        return ws

    async def _subscribe_channels(self, websocket_assistant: WSAssistant):
        """
        Subscribes to order events and balance events.

        :param ws: the websocket assistant used to connect to the exchange
        """
        try:
            orders_change_payload = {
                "id": web_utils.next_message_id(),
                "type": "subscribe",
                "topic": "/spotMarket/tradeOrders",
                "privateChannel": True,
                "response": False,
            }
            subscribe_order_change_request: WSJSONRequest = WSJSONRequest(payload=orders_change_payload)

            balance_payload = {
                "id": web_utils.next_message_id(),
                "type": "subscribe",
                "topic": "/account/balance",
                "privateChannel": True,
                "response": False,
            }
            subscribe_balance_request: WSJSONRequest = WSJSONRequest(payload=balance_payload)

            await websocket_assistant.send(subscribe_order_change_request)
            await websocket_assistant.send(subscribe_balance_request)

            self._last_ws_message_sent_timestamp = self._time()
            self.logger().info("Subscribed to private order changes and balance updates channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to user streams...")
            raise

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant, queue: asyncio.Queue):
        while True:
            try:
                seconds_until_next_ping = self._ping_interval - (self._time() - self._last_ws_message_sent_timestamp)
                await asyncio.wait_for(
                    super()._process_websocket_messages(
                        websocket_assistant=websocket_assistant, queue=queue),
                    timeout=seconds_until_next_ping)
            except asyncio.TimeoutError:
                payload = {
                    "id": web_utils.next_message_id(),
                    "type": "ping",
                }
                ping_request = WSJSONRequest(payload=payload)
                self._last_ws_message_sent_timestamp = self._time()
                await websocket_assistant.send(request=ping_request)

    async def _process_event_message(self, event_message: Dict[str, Any], queue: asyncio.Queue):
        if (len(event_message) > 0
                and event_message.get("type") == "message"
                and event_message.get("subject") in [CONSTANTS.ORDER_CHANGE_EVENT_TYPE, CONSTANTS.BALANCE_EVENT_TYPE]):
            queue.put_nowait(event_message)



import asyncio
from typing import Any, Dict, Optional

from hummingbot.connector.exchange.dexalot import dexalot_constants as CONSTANTS, dexalot_web_utils as web_utils
from hummingbot.connector.exchange.dexalot.dexalot_auth import DexalotAuth
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger


class DexalotAPIUserStreamDataSource(UserStreamTrackerDataSource):
    _logger: Optional[HummingbotLogger] = None

    def __init__(self,
                 auth: DexalotAuth,
                 api_factory: WebAssistantsFactory,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN):
        super().__init__()
        self._auth: DexalotAuth = auth
        self._domain = domain
        self._api_factory = api_factory

    async def _connected_websocket_assistant(self) -> WSAssistant:
        url = f"{web_utils.wss_url(self._domain)}"
        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=url, ping_timeout=CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
        return ws

    async def _subscribe_channels(self, websocket_assistant: WSAssistant):
        """
        Subscribes to order events and balance events.

        :param websocket_assistant: the websocket assistant used to connect to the exchange
        """
        try:

            user_payload = {"type": "tradereventsubscribe"}
            subscribe_order_change_request: WSJSONRequest = WSJSONRequest(payload=user_payload, is_auth_required=True)
            await websocket_assistant.send(subscribe_order_change_request)
            self.logger().info("Subscribed to private order changes and trade updates channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().exception("Unexpected error occurred subscribing to user streams...")
            raise

    async def _process_event_message(self, event_message: Dict[str, Any], queue: asyncio.Queue):
        if event_message.get("type") in [
            CONSTANTS.USER_TRADES_ENDPOINT_NAME,
            CONSTANTS.USER_ORDERS_ENDPOINT_NAME,
        ]:
            queue.put_nowait(event_message)



from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit
from hummingbot.core.data_type.in_flight_order import OrderState

DEFAULT_DOMAIN = "exchange"

REST_URL = "https://api.backpack.{}/"
WSS_URL = "wss://ws.backpack.{}/"


WS_HEARTBEAT_TIME_INTERVAL = 60
MAX_ORDER_ID_LEN = 32  # Full uint32 bit space
HBOT_ORDER_ID_PREFIX = ""  # No prefix - use full ID space for uniqueness
BROKER_ID = 2200

ALL_ORDERS_CHANNEL = "account.orderUpdate"
SINGLE_ORDERS_CHANNEL = "account.orderUpdate.{}"  # format by symbol
ALL_POSITIONS_CHANNEL = "account.positionUpdate"
SINGLE_POSITIONS_CHANNEL = "account.positionUpdate.{}"  # format by symbol

SIDE_BUY = "Bid"
SIDE_SELL = "Ask"
TIME_IN_FORCE_GTC = "GTC"
ORDER_STATE = {
    "Cancelled": OrderState.CANCELED,
    "Expired": OrderState.CANCELED,
    "Filled": OrderState.FILLED,
    "New": OrderState.OPEN,
    "PartiallyFilled": OrderState.PARTIALLY_FILLED,
    "TriggerPending": OrderState.PENDING_CREATE,
    "TriggerFailed": OrderState.FAILED,
}

DIFF_EVENT_TYPE = "depth"
TRADE_EVENT_TYPE = "trade"
FUNDING_EVENT_TYPE = "markPrice"

PING_PATH_URL = "api/v1/ping"
SERVER_TIME_PATH_URL = "api/v1/time"
EXCHANGE_INFO_PATH_URL = "api/v1/markets"
SNAPSHOT_PATH_URL = "api/v1/depth"
BALANCE_PATH_URL = "api/v1/capital"  # instruction balanceQuery
TICKER_BOOK_PATH_URL = "api/v1/tickers"
TICKER_PRICE_CHANGE_PATH_URL = "api/v1/ticker"
ORDER_PATH_URL = "api/v1/order"
MY_TRADES_PATH_URL = "wapi/v1/history/fills"
POSITIONS_PATH_URL = "api/v1/position"
MARK_PRICE_PATH_URL = "api/v1/markPrices"
FUNDING_RATE_PATH_URL = "api/v1/fundingRates"
FUNDING_PAYMENTS_PATH_URL = "wapi/v1/history/funding"
ACCOUNT_PATH_URL = "api/v1/account"


GLOBAL_RATE_LIMIT = "GLOBAL"

# Present in https://support.backpack.exchange/exchange/api-and-developer-docs/faqs, not in the docs
RATE_LIMITS = [
    # Global pool limit
    RateLimit(limit_id=GLOBAL_RATE_LIMIT, limit=2000, time_interval=60),
    # All endpoints linked to the global pool
    RateLimit(
        limit_id=SERVER_TIME_PATH_URL,
        limit=2000,
        time_interval=60,
        linked_limits=[LinkedLimitWeightPair(GLOBAL_RATE_LIMIT)],
    ),
    RateLimit(
        limit_id=EXCHANGE_INFO_PATH_URL,
        limit=2000,
        time_interval=60,
        linked_limits=[LinkedLimitWeightPair(GLOBAL_RATE_LIMIT)],
    ),
    RateLimit(
        limit_id=PING_PATH_URL,
        limit=2000,
        time_interval=60,
        linked_limits=[LinkedLimitWeightPair(GLOBAL_RATE_LIMIT)],
    ),
    RateLimit(
        limit_id=SNAPSHOT_PATH_URL,
        limit=2000,
        time_interval=60,
        linked_limits=[LinkedLimitWeightPair(GLOBAL_RATE_LIMIT)],
    ),
    RateLimit(
        limit_id=BALANCE_PATH_URL,
        limit=2000,
        time_interval=60,
        linked_limits=[LinkedLimitWeightPair(GLOBAL_RATE_LIMIT)],
    ),
    RateLimit(
        limit_id=TICKER_BOOK_PATH_URL,
        limit=2000,
        time_interval=60,
        linked_limits=[LinkedLimitWeightPair(GLOBAL_RATE_LIMIT)],
    ),
    RateLimit(
        limit_id=TICKER_PRICE_CHANGE_PATH_URL,
        limit=2000,
        time_interval=60,
        linked_limits=[LinkedLimitWeightPair(GLOBAL_RATE_LIMIT)],
    ),
    RateLimit(
        limit_id=ORDER_PATH_URL,
        limit=2000,
        time_interval=60,
        linked_limits=[LinkedLimitWeightPair(GLOBAL_RATE_LIMIT)],
    ),
    RateLimit(
        limit_id=MY_TRADES_PATH_URL,
        limit=2000,
        time_interval=60,
        linked_limits=[LinkedLimitWeightPair(GLOBAL_RATE_LIMIT)],
    ),
    RateLimit(
        limit_id=POSITIONS_PATH_URL,
        limit=2000,
        time_interval=60,
        linked_limits=[LinkedLimitWeightPair(GLOBAL_RATE_LIMIT)],
    ),
    RateLimit(
        limit_id=FUNDING_PAYMENTS_PATH_URL,
        limit=2000,
        time_interval=60,
        linked_limits=[LinkedLimitWeightPair(GLOBAL_RATE_LIMIT)],
    ),
    RateLimit(
        limit_id=MARK_PRICE_PATH_URL,
        limit=2000,
        time_interval=60,
        linked_limits=[LinkedLimitWeightPair(GLOBAL_RATE_LIMIT)],
    ),
    RateLimit(
        limit_id=ACCOUNT_PATH_URL,
        limit=2000,
        time_interval=60,
        linked_limits=[LinkedLimitWeightPair(GLOBAL_RATE_LIMIT)],
    ),
]

ORDER_NOT_EXIST_ERROR_CODE = "RESOURCE_NOT_FOUND"
ORDER_NOT_EXIST_MESSAGE = "Not Found"
UNKNOWN_ORDER_ERROR_CODE = "RESOURCE_NOT_FOUND"
UNKNOWN_ORDER_MESSAGE = "Not Found"



import hashlib
import hmac
from datetime import datetime, timezone
from typing import Dict

from hummingbot.connector.exchange.foxbit import foxbit_web_utils as web_utils
from hummingbot.connector.time_synchronizer import TimeSynchronizer
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest, WSRequest


class FoxbitAuth(AuthBase):

    def __init__(self, api_key: str, secret_key: str, user_id: str, time_provider: TimeSynchronizer):
        self.api_key = api_key
        self.secret_key = secret_key
        self.user_id = user_id
        self.time_provider = time_provider

    async def rest_authenticate(self,
                                request: RESTRequest,
                                ) -> RESTRequest:
        """
        Adds the server time and the signature to the request, required for authenticated interactions. It also adds
        the required parameter in the request header.
        :param request: the request to be configured for authenticated interaction
        """
        timestamp = str(int(datetime.now(timezone.utc).timestamp() * 1e3))

        endpoint_url = web_utils.rest_endpoint_url(request.url)

        params = request.params if request.params is not None else ""
        if request.method == RESTMethod.GET and request.params is not None:
            params = ''
            i = 0
            for p in request.params:
                k = p
                v = request.params[p]
                if i == 0:
                    params = params + f"{k}={v}"
                else:
                    params = params + f"&{k}={v}"
                i += 1

        data = request.data if request.data is not None else ""

        to_payload = params if len(params) > 0 else data

        payload = '{}{}{}{}'.format(timestamp,
                                    request.method,
                                    endpoint_url,
                                    to_payload
                                    )

        signature = hmac.new(self.secret_key.encode("utf8"),
                             payload.encode("utf8"),
                             hashlib.sha256).digest().hex()

        foxbit_header = {
            "X-FB-ACCESS-KEY": self.api_key,
            "X-FB-ACCESS-SIGNATURE": signature,
            "X-FB-ACCESS-TIMESTAMP": timestamp,
        }

        headers = {}
        if request.headers is not None:
            headers.update(request.headers)
        headers.update(foxbit_header)
        request.headers = headers

        return request

    async def ws_authenticate(self,
                              request: WSRequest,
                              ) -> WSRequest:
        """
        This method is intended to configure a websocket request to be authenticated.
        It should be used with empty requests to send an initial login payload.
        :param request: the request to be configured for authenticated interaction
        """

        request.payload = self.get_ws_authenticate_payload(request)
        return request

    def get_ws_authenticate_payload(self,
                                    request: WSRequest = None,
                                    ) -> Dict[str, any]:
        timestamp = int(datetime.now(timezone.utc).timestamp() * 1e3)

        msg = '{}{}{}'.format(timestamp,
                              self.user_id,
                              self.api_key)

        signature = hmac.new(self.secret_key.encode("utf8"),
                             msg.encode("utf8"),
                             hashlib.sha256).digest().hex()

        payload = {
            "APIKey": self.api_key,
            "Signature": signature,
            "UserId": self.user_id,
            "Nonce": timestamp
        }

        if hasattr(request, 'payload'):
            payload.update(request.payload)

        return payload



from typing import List, NamedTuple

import pandas as pd

from hummingbot.core.data_type.common import OrderType, PositionAction


class MarketOrder(NamedTuple):
    order_id: str
    trading_pair: str
    is_buy: bool
    base_asset: str
    quote_asset: str
    amount: float
    timestamp: float
    position: PositionAction = PositionAction.NIL

    @classmethod
    def to_pandas(cls, market_orders: List["MarketOrder"]) -> pd.DataFrame:
        columns = ["order_id", "trading_pair", "is_buy", "base_asset", "quote_asset", "quantity", "timestamp"]
        data = [[
            market_order.order_id,
            market_order.trading_pair,
            market_order.is_buy,
            market_order.base_asset,
            market_order.quote_asset,
            market_order.amount,
            pd.Timestamp(market_order.timestamp, unit='s', tz='UTC').strftime('%Y-%m-%d %H:%M:%S')
        ] for market_order in market_orders]
        return pd.DataFrame(data=data, columns=columns)

    @property
    def client_order_id(self):
        # Added to make this class polymorphic with LimitOrder
        return self.order_id

    @property
    def quantity(self):
        # Added to make this class polymorphic with LimitOrder
        return self.amount

    @property
    def price(self):
        # Added to make this class polymorphic with LimitOrder
        return None

    def order_type(self) -> OrderType:
        return OrderType.MARKET

    def copy_with_id(self, client_order_id: str):
        return MarketOrder(
            order_id=client_order_id,
            trading_pair=self.trading_pair,
            is_buy=self.is_buy,
            base_asset=self.base_asset,
            quote_asset=self.quote_asset,
            amount=self.amount,
            timestamp=self.timestamp,
            position=self.position,
        )



from decimal import Decimal
from typing import TYPE_CHECKING, Dict, Optional

from hummingbot.core.rate_oracle.sources.rate_source_base import RateSourceBase
from hummingbot.core.utils import async_ttl_cache

if TYPE_CHECKING:
    from hummingbot.connector.exchange.dexalot.dexalot_exchange import DexalotExchange


class DexalotRateSource(RateSourceBase):
    def __init__(self):
        super().__init__()
        self._exchange: Optional[DexalotExchange] = None  # delayed because of circular reference

    @property
    def name(self) -> str:
        return "dexalot"

    @async_ttl_cache(ttl=30, maxsize=1)
    async def get_prices(self, quote_token: Optional[str] = None) -> Dict[str, Decimal]:
        self._ensure_exchange()
        results = {}
        try:
            records = await self._exchange.get_all_pairs_prices()
            for record in records:
                try:
                    pair = await self._exchange.trading_pair_associated_to_exchange_symbol(record["pair"])
                except KeyError:
                    # Ignore results for which their symbols is not tracked by the connector
                    continue

                if Decimal(str(record["low"])) > 0 and Decimal(str(record["high"])) > 0:
                    results[pair] = (Decimal(str(record["low"])) +
                                     Decimal(str(record["high"]))) / Decimal("2")
        except Exception:
            self.logger().exception(
                msg="Unexpected error while retrieving rates from Dexalot. Check the log file for more info.",
            )
        return results

    def _ensure_exchange(self):
        if self._exchange is None:
            self._exchange = self._build_dexalot_connector_without_private_keys()

    @staticmethod
    def _build_dexalot_connector_without_private_keys() -> 'DexalotExchange':
        from hummingbot.connector.exchange.dexalot.dexalot_exchange import DexalotExchange

        return DexalotExchange(
            dexalot_api_key="",
            dexalot_api_secret="13e56ca9cceebf1f33065c2c5376ab38570a114bc1b003b60d838f92be9d7930",  # noqa: mock
            trading_pairs=[],
            trading_required=False,
        )



from decimal import Decimal
import logging
import pandas as pd
from typing import (
    List)

from hummingbot.core.clock cimport Clock
from hummingbot.core.event.events import MarketEvent, AccountEvent
from hummingbot.core.event.event_listener cimport EventListener
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple
from hummingbot.core.time_iterator cimport TimeIterator
from hummingbot.connector.connector_base cimport ConnectorBase
from hummingbot.core.data_type.trade import Trade
from hummingbot.core.event.events import OrderFilledEvent
from hummingbot.core.data_type.common import OrderType, PositionAction
from hummingbot.strategy.order_tracker import OrderTracker
from hummingbot.connector.derivative_base import DerivativeBase

NaN = float("nan")
s_decimal_nan = Decimal("NaN")
s_decimal_0 = Decimal("0")

# <editor-fold desc="+ Event listeners">
cdef class BaseStrategyEventListener(EventListener):
    cdef:
        StrategyBase _owner

    def __init__(self, StrategyBase owner):
        super().__init__()
        self._owner = owner


cdef class BuyOrderCompletedListener(BaseStrategyEventListener):
    cdef c_call(self, object arg):
        self._owner.c_did_complete_buy_order(arg)
        self._owner.c_did_complete_buy_order_tracker(arg)


cdef class SellOrderCompletedListener(BaseStrategyEventListener):
    cdef c_call(self, object arg):
        self._owner.c_did_complete_sell_order(arg)
        self._owner.c_did_complete_sell_order_tracker(arg)


cdef class FundingPaymentCompletedListener(BaseStrategyEventListener):
    cdef c_call(self, object arg):
        self._owner.c_did_complete_funding_payment(arg)


cdef class PositionModeChangeSuccessListener(BaseStrategyEventListener):
    cdef c_call(self, object arg):
        self._owner.c_did_change_position_mode_succeed(arg)

cdef class PositionModeChangeFailureListener(BaseStrategyEventListener):
    cdef c_call(self, object arg):
        self._owner.c_did_change_position_mode_fail(arg)

cdef class OrderFilledListener(BaseStrategyEventListener):
    cdef c_call(self, object arg):
        self._owner.c_did_fill_order(arg)


cdef class OrderFailedListener(BaseStrategyEventListener):
    cdef c_call(self, object arg):
        self._owner.c_did_fail_order(arg)
        self._owner.c_did_fail_order_tracker(arg)


cdef class OrderCancelledListener(BaseStrategyEventListener):
    cdef c_call(self, object arg):
        self._owner.c_did_cancel_order(arg)
        self._owner.c_did_cancel_order_tracker(arg)


cdef class OrderExpiredListener(BaseStrategyEventListener):
    cdef c_call(self, object arg):
        self._owner.c_did_expire_order(arg)
        self._owner.c_did_expire_order_tracker(arg)


cdef class BuyOrderCreatedListener(BaseStrategyEventListener):
    cdef c_call(self, object arg):
        self._owner.c_did_create_buy_order(arg)


cdef class SellOrderCreatedListener(BaseStrategyEventListener):
    cdef c_call(self, object arg):
        self._owner.c_did_create_sell_order(arg)

cdef class RangePositionLiquidityAddedListener(BaseStrategyEventListener):
    cdef c_call(self, object arg):
        self._owner.c_did_add_liquidity(arg)

cdef class RangePositionLiquidityRemovedListener(BaseStrategyEventListener):
    cdef c_call(self, object arg):
        self._owner.c_did_remove_liquidity(arg)

cdef class RangePositionUpdateFailureListener(BaseStrategyEventListener):
    cdef c_call(self, object arg):
        self._owner.c_did_fail_lp_update(arg)
# </editor-fold>


cdef class StrategyBase(TimeIterator):
    BUY_ORDER_COMPLETED_EVENT_TAG = MarketEvent.BuyOrderCompleted.value
    SELL_ORDER_COMPLETED_EVENT_TAG = MarketEvent.SellOrderCompleted.value
    FUNDING_PAYMENT_COMPLETED_EVENT_TAG = MarketEvent.FundingPaymentCompleted.value
    POSITION_MODE_CHANGE_SUCCEEDED_EVENT_TAG = AccountEvent.PositionModeChangeSucceeded.value
    POSITION_MODE_CHANGE_FAILED_EVENT_TAG = AccountEvent.PositionModeChangeFailed.value
    ORDER_FILLED_EVENT_TAG = MarketEvent.OrderFilled.value
    ORDER_CANCELED_EVENT_TAG = MarketEvent.OrderCancelled.value
    ORDER_EXPIRED_EVENT_TAG = MarketEvent.OrderExpired.value
    ORDER_FAILURE_EVENT_TAG = MarketEvent.OrderFailure.value
    BUY_ORDER_CREATED_EVENT_TAG = MarketEvent.BuyOrderCreated.value
    SELL_ORDER_CREATED_EVENT_TAG = MarketEvent.SellOrderCreated.value
    RANGE_POSITION_LIQUIDITY_ADDED_EVENT_TAG = MarketEvent.RangePositionLiquidityAdded.value
    RANGE_POSITION_LIQUIDITY_REMOVED_EVENT_TAG = MarketEvent.RangePositionLiquidityRemoved.value
    RANGE_POSITION_UPDATE_FAILURE_EVENT_TAG = MarketEvent.RangePositionUpdateFailure.value


    @classmethod
    def logger(cls) -> logging.Logger:
        raise NotImplementedError

    def __init__(self):
        super().__init__()
        self._sb_markets = set()
        self._sb_create_buy_order_listener = BuyOrderCreatedListener(self)
        self._sb_create_sell_order_listener = SellOrderCreatedListener(self)
        self._sb_fill_order_listener = OrderFilledListener(self)
        self._sb_fail_order_listener = OrderFailedListener(self)
        self._sb_cancel_order_listener = OrderCancelledListener(self)
        self._sb_expire_order_listener = OrderExpiredListener(self)
        self._sb_complete_buy_order_listener = BuyOrderCompletedListener(self)
        self._sb_complete_sell_order_listener = SellOrderCompletedListener(self)
        self._sb_complete_funding_payment_listener = FundingPaymentCompletedListener(self)
        self._sb_position_mode_change_success_listener = PositionModeChangeSuccessListener(self)
        self._sb_position_mode_change_failure_listener = PositionModeChangeFailureListener(self)
        self._sb_range_position_liquidity_added_listener = RangePositionLiquidityAddedListener(self)
        self._sb_range_position_liquidity_removed_listener = RangePositionLiquidityRemovedListener(self)
        self._sb_range_position_update_failure_listener = RangePositionUpdateFailureListener(self)

        self._sb_delegate_lock = False

        self._sb_order_tracker = OrderTracker()

    def init_params(self, *args, **kwargs):
        """
        Assigns strategy parameters, this function must be called directly after init.
        The reason for this is to make the parameters discoverable through introspect (this is not possible on init of
        a Cython class).
        """
        raise NotImplementedError

    @property
    def active_markets(self) -> List[ConnectorBase]:
        return list(self._sb_markets)

    @property
    def order_tracker(self) -> OrderTracker:
        return self._sb_order_tracker

    def format_status(self):
        raise NotImplementedError

    def log_with_clock(self, log_level: int, msg: str, **kwargs):
        clock_timestamp = pd.Timestamp(self._current_timestamp, unit="s", tz="UTC")
        self.logger().log(log_level, f"{msg} [clock={str(clock_timestamp)}]", **kwargs)

    @property
    def trades(self) -> List[Trade]:
        """
        Returns a list of all completed trades from the market.
        The trades are taken from the market event logs.
        """
        def event_to_trade(order_filled_event: OrderFilledEvent, market_name: str):
            return Trade(order_filled_event.trading_pair,
                         order_filled_event.trade_type,
                         order_filled_event.price,
                         order_filled_event.amount,
                         order_filled_event.order_type,
                         market_name,
                         order_filled_event.timestamp,
                         order_filled_event.trade_fee)
        past_trades = []
        for market in self.active_markets:
            event_logs = market.event_logs
            order_filled_events = list(filter(lambda e: isinstance(e, OrderFilledEvent), event_logs))
            past_trades += list(map(lambda ofe: event_to_trade(ofe, market.display_name), order_filled_events))

        return sorted(past_trades, key=lambda x: x.timestamp)

    def market_status_data_frame(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> pd.DataFrame:
        cdef:
            ConnectorBase market
            str trading_pair
            str base_asset
            str quote_asset
            object bid_price
            object ask_price
            list markets_data = []
            list markets_columns = ["Exchange", "Market", "Best Bid Price", "Best Ask Price", "Mid Price"]
        try:
            for market_trading_pair_tuple in market_trading_pair_tuples:
                market, trading_pair, base_asset, quote_asset = market_trading_pair_tuple
                bid_price = market.get_price(trading_pair, False)
                ask_price = market.get_price(trading_pair, True)
                mid_price = (bid_price + ask_price)/2
                markets_data.append([
                    market.display_name,
                    trading_pair,
                    float(bid_price),
                    float(ask_price),
                    float(mid_price)
                ])
            return pd.DataFrame(data=markets_data, columns=markets_columns)

        except Exception:
            self.logger().error("Error formatting market stats.", exc_info=True)

    def wallet_balance_data_frame(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> pd.DataFrame:
        cdef:
            ConnectorBase market
            str base_asset
            str quote_asset
            double base_balance
            double quote_balance
            double base_asset_conversion_rate
            double quote_asset_conversion_rate
            list assets_data = []
            list assets_columns = ["Exchange", "Asset", "Total Balance", "Available Balance"]
        try:
            for market_trading_pair_tuple in market_trading_pair_tuples:
                market, trading_pair, base_asset, quote_asset = market_trading_pair_tuple
                base_balance = float(market.get_balance(base_asset))
                quote_balance = float(market.get_balance(quote_asset))
                available_base_balance = float(market.get_available_balance(base_asset))
                available_quote_balance = float(market.get_available_balance(quote_asset))
                assets_data.extend([
                    [market.display_name, base_asset, base_balance, available_base_balance],
                    [market.display_name, quote_asset, quote_balance, available_quote_balance]
                ])

            return pd.DataFrame(data=assets_data, columns=assets_columns)

        except Exception:
            self.logger().error("Error formatting wallet balance stats.", exc_info=True)

    def balance_warning(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> List[str]:
        cdef:
            double base_balance
            double quote_balance
            list warning_lines = []
        # Add warning lines on null balances.
        # TO-DO: $Use min order size logic to replace the hard-coded 0.0001 value for each asset.
        for market_trading_pair_tuple in market_trading_pair_tuples:
            base_balance = market_trading_pair_tuple.market.get_balance(market_trading_pair_tuple.base_asset)
            quote_balance = market_trading_pair_tuple.market.get_balance(market_trading_pair_tuple.quote_asset)
            if base_balance <= Decimal("0.0001") and not isinstance(market_trading_pair_tuple.market, DerivativeBase):
                warning_lines.append(f"  {market_trading_pair_tuple.market.name} market "
                                     f"{market_trading_pair_tuple.base_asset} balance is too low. Cannot place order.")
            if quote_balance <= Decimal("0.0001"):
                warning_lines.append(f"  {market_trading_pair_tuple.market.name} market "
                                     f"{market_trading_pair_tuple.quote_asset} balance is too low. Cannot place order.")
        return warning_lines

    def network_warning(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> List[str]:
        cdef:
            list warning_lines = []
            str trading_pairs
        if not all([market_trading_pair_tuple.market.network_status is NetworkStatus.CONNECTED for
                    market_trading_pair_tuple in market_trading_pair_tuples]):
            trading_pairs = " // ".join([market_trading_pair_tuple.trading_pair for market_trading_pair_tuple in market_trading_pair_tuples])
            warning_lines.extend([
                f"  Markets are offline for the {trading_pairs} pair. Continued trading "
                f"with these markets may be dangerous.",
                ""
            ])
        return warning_lines

    cdef c_start(self, Clock clock, double timestamp):
        TimeIterator.c_start(self, clock, timestamp)
        self._sb_order_tracker.c_start(clock, timestamp)

    cdef c_tick(self, double timestamp):
        TimeIterator.c_tick(self, timestamp)
        self._sb_order_tracker.c_tick(timestamp)

    cdef c_stop(self, Clock clock):
        TimeIterator.c_stop(self, clock)
        self._sb_order_tracker.c_stop(clock)
        self.c_remove_markets(list(self._sb_markets))

    cdef c_add_markets(self, list markets):
        cdef:
            ConnectorBase typed_market

        for market in markets:
            typed_market = market
            typed_market.c_add_listener(self.BUY_ORDER_CREATED_EVENT_TAG, self._sb_create_buy_order_listener)
            typed_market.c_add_listener(self.SELL_ORDER_CREATED_EVENT_TAG, self._sb_create_sell_order_listener)
            typed_market.c_add_listener(self.ORDER_FILLED_EVENT_TAG, self._sb_fill_order_listener)
            typed_market.c_add_listener(self.ORDER_FAILURE_EVENT_TAG, self._sb_fail_order_listener)
            typed_market.c_add_listener(self.ORDER_CANCELED_EVENT_TAG, self._sb_cancel_order_listener)
            typed_market.c_add_listener(self.ORDER_EXPIRED_EVENT_TAG, self._sb_expire_order_listener)
            typed_market.c_add_listener(self.BUY_ORDER_COMPLETED_EVENT_TAG, self._sb_complete_buy_order_listener)
            typed_market.c_add_listener(self.SELL_ORDER_COMPLETED_EVENT_TAG, self._sb_complete_sell_order_listener)
            typed_market.c_add_listener(self.FUNDING_PAYMENT_COMPLETED_EVENT_TAG, self._sb_complete_funding_payment_listener)
            typed_market.c_add_listener(self.POSITION_MODE_CHANGE_SUCCEEDED_EVENT_TAG, self._sb_position_mode_change_success_listener)
            typed_market.c_add_listener(self.POSITION_MODE_CHANGE_FAILED_EVENT_TAG, self._sb_position_mode_change_failure_listener)
            typed_market.c_add_listener(self.RANGE_POSITION_LIQUIDITY_ADDED_EVENT_TAG, self._sb_range_position_liquidity_added_listener)
            typed_market.c_add_listener(self.RANGE_POSITION_LIQUIDITY_REMOVED_EVENT_TAG, self._sb_range_position_liquidity_removed_listener)
            typed_market.c_add_listener(self.RANGE_POSITION_UPDATE_FAILURE_EVENT_TAG, self._sb_range_position_update_failure_listener)
            self._sb_markets.add(typed_market)

    def add_markets(self, markets: List[ConnectorBase]):
        self.c_add_markets(markets)

    cdef c_remove_markets(self, list markets):
        cdef:
            ConnectorBase typed_market

        for market in markets:
            typed_market = market
            if typed_market not in self._sb_markets:
                continue
            typed_market.c_remove_listener(self.BUY_ORDER_CREATED_EVENT_TAG, self._sb_create_buy_order_listener)
            typed_market.c_remove_listener(self.SELL_ORDER_CREATED_EVENT_TAG, self._sb_create_sell_order_listener)
            typed_market.c_remove_listener(self.ORDER_FILLED_EVENT_TAG, self._sb_fill_order_listener)
            typed_market.c_remove_listener(self.ORDER_FAILURE_EVENT_TAG, self._sb_fail_order_listener)
            typed_market.c_remove_listener(self.ORDER_CANCELED_EVENT_TAG, self._sb_cancel_order_listener)
            typed_market.c_remove_listener(self.ORDER_EXPIRED_EVENT_TAG, self._sb_expire_order_listener)
            typed_market.c_remove_listener(self.BUY_ORDER_COMPLETED_EVENT_TAG, self._sb_complete_buy_order_listener)
            typed_market.c_remove_listener(self.SELL_ORDER_COMPLETED_EVENT_TAG, self._sb_complete_sell_order_listener)
            typed_market.c_remove_listener(self.FUNDING_PAYMENT_COMPLETED_EVENT_TAG, self._sb_complete_funding_payment_listener)
            typed_market.c_remove_listener(self.POSITION_MODE_CHANGE_SUCCEEDED_EVENT_TAG, self._sb_position_mode_change_success_listener)
            typed_market.c_remove_listener(self.POSITION_MODE_CHANGE_FAILED_EVENT_TAG, self._sb_position_mode_change_failure_listener)
            typed_market.c_remove_listener(self.RANGE_POSITION_LIQUIDITY_ADDED_EVENT_TAG, self._sb_range_position_liquidity_added_listener)
            typed_market.c_remove_listener(self.RANGE_POSITION_LIQUIDITY_REMOVED_EVENT_TAG, self._sb_range_position_liquidity_removed_listener)
            typed_market.c_remove_listener(self.RANGE_POSITION_UPDATE_FAILURE_EVENT_TAG, self._sb_range_position_update_failure_listener)
            self._sb_markets.remove(typed_market)

    def remove_markets(self, markets: List[ConnectorBase]):
        self.c_remove_markets(markets)

    cdef object c_sum_flat_fees(self, str quote_asset, list flat_fees):

        """
        Converts flat fees to quote token and sums up all flat fees
        """
        cdef:
            object total_flat_fees = s_decimal_0

        for flat_fee_currency, flat_fee_amount in flat_fees:
            if flat_fee_currency == quote_asset:
                total_flat_fees += flat_fee_amount
            else:
                # if the flat fee currency asset does not match quote asset, raise exception for now
                # as we don't support different token conversion atm.
                raise Exception("Flat fee in other token than quote asset is not supported.")
        return total_flat_fees

    def cum_flat_fees(self, quote_asset: str, flat_fees: List):
        return self.c_sum_flat_fees(quote_asset, flat_fees)

    # <editor-fold desc="+ Market event interfaces">
    # ----------------------------------------------------------------------------------------------------------
    cdef c_did_create_buy_order(self, object order_created_event):
        """
        In the case of asynchronous order creation on the exchange's server, this event is NOT triggered
        upon submission of the order request to the server - it is only triggered once the server has sent
        the acknowledgment that the order is successfully created.
        """
        pass

    cdef c_did_create_sell_order(self, object order_created_event):
        """
        In the case of asynchronous order creation on the exchange's server, this event is NOT triggered
        upon submission of the order request to the server - it is only triggered once the server has sent
        the acknowledgment that the order is successfully created.
        """
        pass

    cdef c_did_fill_order(self, object order_filled_event):
        pass

    cdef c_did_fail_order(self, object order_failed_event):
        pass

    cdef c_did_cancel_order(self, object cancelled_event):
        pass

    cdef c_did_expire_order(self, object expired_event):
        pass

    cdef c_did_complete_buy_order(self, object order_completed_event):
        pass

    cdef c_did_complete_sell_order(self, object order_completed_event):
        pass

    cdef c_did_complete_funding_payment(self, object funding_payment_completed_event):
        pass

    cdef c_did_change_position_mode_succeed(self, object position_mode_changed_event):
        pass

    cdef c_did_change_position_mode_fail(self, object position_mode_changed_event):
        pass

    cdef c_did_add_liquidity(self, object add_liquidity_event):
        pass

    cdef c_did_remove_liquidity(self, object remove_liquidity_event):
        pass

    cdef c_did_fail_lp_update(self, object fail_lp_update_event):
        pass
    # ----------------------------------------------------------------------------------------------------------
    # </editor-fold>

    # <editor-fold desc="+ Order tracking event handlers">
    # ----------------------------------------------------------------------------------------------------------
    cdef c_did_fail_order_tracker(self, object order_failed_event):
        cdef:
            str order_id = order_failed_event.order_id
            object order_type = order_failed_event.order_type
            object market_pair = self._sb_order_tracker.c_get_market_pair_from_order_id(order_id)

        if order_type.is_limit_type():
            self.c_stop_tracking_limit_order(market_pair, order_id)
        elif order_type == OrderType.MARKET:
            self.c_stop_tracking_market_order(market_pair, order_id)

    cdef c_did_cancel_order_tracker(self, object order_cancelled_event):
        cdef:
            str order_id = order_cancelled_event.order_id
            object market_pair = self._sb_order_tracker.c_get_market_pair_from_order_id(order_id)

        self.c_stop_tracking_limit_order(market_pair, order_id)

    cdef c_did_expire_order_tracker(self, object order_expired_event):
        self.c_did_cancel_order_tracker(order_expired_event)

    cdef c_did_complete_buy_order_tracker(self, object order_completed_event):
        cdef:
            str order_id = order_completed_event.order_id
            object market_pair = self._sb_order_tracker.c_get_market_pair_from_order_id(order_id)
            object order_type = order_completed_event.order_type

        if market_pair is not None:
            if order_type.is_limit_type():
                self.c_stop_tracking_limit_order(market_pair, order_id)
            elif order_type == OrderType.MARKET:
                self.c_stop_tracking_market_order(market_pair, order_id)

    cdef c_did_complete_sell_order_tracker(self, object order_completed_event):
        self.c_did_complete_buy_order_tracker(order_completed_event)

    # ----------------------------------------------------------------------------------------------------------
    # </editor-fold>

    # <editor-fold desc="+ Creating and canceling orders">
    # ----------------------------------------------------------------------------------------------------------

    def buy_with_specific_market(self, market_trading_pair_tuple, amount,
                                 order_type=OrderType.MARKET,
                                 price=s_decimal_nan,
                                 expiration_seconds=NaN,
                                 position_action=PositionAction.OPEN):
        return self.c_buy_with_specific_market(market_trading_pair_tuple, amount,
                                               order_type,
                                               price,
                                               expiration_seconds,
                                               position_action)

    cdef str c_buy_with_specific_market(self, object market_trading_pair_tuple, object amount,
                                        object order_type=OrderType.MARKET,
                                        object price=s_decimal_nan,
                                        double expiration_seconds=NaN,
                                        position_action=PositionAction.OPEN):
        if self._sb_delegate_lock:
            raise RuntimeError("Delegates are not allowed to execute orders directly.")

        if not (isinstance(amount, Decimal) and isinstance(price, Decimal)):
            raise TypeError("price and amount must be Decimal objects.")

        cdef:
            kwargs = {"expiration_ts": self._current_timestamp + expiration_seconds,
                      "position_action": position_action}
            ConnectorBase market = market_trading_pair_tuple.market

        if market not in self._sb_markets:
            raise ValueError(f"Market object for buy order is not in the whitelisted markets set.")

        cdef:
            str order_id = market.c_buy(market_trading_pair_tuple.trading_pair,
                                        amount=amount,
                                        order_type=order_type,
                                        price=price,
                                        kwargs=kwargs)

        # Start order tracking
        if order_type.is_limit_type():
            self.c_start_tracking_limit_order(market_trading_pair_tuple, order_id, True, price, amount)
        elif order_type == OrderType.MARKET:
            self.c_start_tracking_market_order(market_trading_pair_tuple, order_id, True, amount)

        return order_id

    def sell_with_specific_market(self, market_trading_pair_tuple, amount,
                                  order_type=OrderType.MARKET,
                                  price=s_decimal_nan,
                                  expiration_seconds=NaN,
                                  position_action=PositionAction.OPEN):
        return self.c_sell_with_specific_market(market_trading_pair_tuple, amount,
                                                order_type,
                                                price,
                                                expiration_seconds,
                                                position_action)

    cdef str c_sell_with_specific_market(self, object market_trading_pair_tuple, object amount,
                                         object order_type=OrderType.MARKET,
                                         object price=s_decimal_nan,
                                         double expiration_seconds=NaN,
                                         position_action=PositionAction.OPEN):
        if self._sb_delegate_lock:
            raise RuntimeError("Delegates are not allowed to execute orders directly.")

        if not (isinstance(amount, Decimal) and isinstance(price, Decimal)):
            raise TypeError("price and amount must be Decimal objects.")

        cdef:
            kwargs = {"expiration_ts": self._current_timestamp + expiration_seconds,
                      "position_action": position_action}
            ConnectorBase market = market_trading_pair_tuple.market

        if market not in self._sb_markets:
            raise ValueError(f"Market object for sell order is not in the whitelisted markets set.")

        cdef:
            str order_id = market.c_sell(market_trading_pair_tuple.trading_pair, amount,
                                         order_type=order_type, price=price, kwargs=kwargs)

        # Start order tracking
        if order_type.is_limit_type():
            self.c_start_tracking_limit_order(market_trading_pair_tuple, order_id, False, price, amount)
        elif order_type == OrderType.MARKET:
            self.c_start_tracking_market_order(market_trading_pair_tuple, order_id, False, amount)

        return order_id

    cdef c_cancel_order(self, object market_trading_pair_tuple, str order_id):
        cdef:
            ConnectorBase market = market_trading_pair_tuple.market

        if self._sb_order_tracker.c_check_and_track_cancel(order_id):
            self.log_with_clock(
                logging.INFO,
                f"({market_trading_pair_tuple.trading_pair}) Canceling the limit order {order_id}."
            )
            market.c_cancel(market_trading_pair_tuple.trading_pair, order_id)

    def cancel_order(self, market_trading_pair_tuple: MarketTradingPairTuple, order_id: str):
        self.c_cancel_order(market_trading_pair_tuple, order_id)
    # ----------------------------------------------------------------------------------------------------------
    # </editor-fold>

    # <editor-fold desc="+ Order tracking entry points">
    # The following exposed tracking functions are meant to allow extending order tracking behavior in strategy
    # classes.
    # ----------------------------------------------------------------------------------------------------------
    cdef c_start_tracking_limit_order(self, object market_pair, str order_id, bint is_buy, object price,
                                      object quantity):
        self._sb_order_tracker.c_start_tracking_limit_order(market_pair, order_id, is_buy, price, quantity)

    def start_tracking_limit_order(self, market_pair: MarketTradingPairTuple, order_id: str, is_buy: bool, price: Decimal,
                                   quantity: Decimal):

        self.c_start_tracking_limit_order(market_pair, order_id, is_buy, price, quantity)

    cdef c_stop_tracking_limit_order(self, object market_pair, str order_id):
        self._sb_order_tracker.c_stop_tracking_limit_order(market_pair, order_id)

    def stop_tracking_limit_order(self, market_pair: MarketTradingPairTuple, order_id: str):
        self.c_stop_tracking_limit_order(market_pair, order_id)

    cdef c_start_tracking_market_order(self, object market_pair, str order_id, bint is_buy, object quantity):
        self._sb_order_tracker.c_start_tracking_market_order(market_pair, order_id, is_buy, quantity)

    def start_tracking_market_order(self, market_pair: MarketTradingPairTuple, order_id: str, is_buy: bool, quantity: Decimal):
        self.c_start_tracking_market_order(market_pair, order_id, is_buy, quantity)

    cdef c_stop_tracking_market_order(self, object market_pair, str order_id):
        self._sb_order_tracker.c_stop_tracking_market_order(market_pair, order_id)

    def stop_tracking_market_order(self, market_pair: MarketTradingPairTuple, order_id: str):
        self.c_stop_tracking_market_order(market_pair, order_id)

    cdef c_track_restored_orders(self, object market_pair):
        cdef:
            list limit_orders = market_pair.market.limit_orders
            list restored_order_ids = []

        for order in limit_orders:
            restored_order_ids.append(order.client_order_id)
            self.c_start_tracking_limit_order(market_pair,
                                              order.client_order_id,
                                              order.is_buy,
                                              order.price,
                                              order.quantity)
        return restored_order_ids

    def track_restored_orders(self, market_pair: MarketTradingPairTuple):
        return self.c_track_restored_orders(market_pair)

    def notify_hb_app(self, msg: str):
        """
        Method called to display message on the Output Panel(upper left)
        :param msg: The message to be notified
        """
        from hummingbot.client.hummingbot_application import HummingbotApplication
        HummingbotApplication.main_application().notify(msg)

    def notify_hb_app_with_timestamp(self, msg: str):
        """
        Method called to display message on the Output Panel(upper left)
        This implementation adds the timestamp as the first element of the notification
        :param msg: The message to be notified
        """
        timestamp = pd.Timestamp.fromtimestamp(self._current_timestamp)
        self.notify_hb_app(f"({timestamp}) {msg}")
    # ----------------------------------------------------------------------------------------------------------
    # </editor-fold>



from decimal import Decimal

from hummingbot.connector.exchange_base import ExchangeBase
from hummingbot.core.data_type.common import PriceType

cdef class AssetPriceDelegate:
    # The following exposed Python functions are meant for unit tests
    # ---------------------------------------------------------------
    def get_mid_price(self) -> Decimal:
        return self.c_get_mid_price()
    # ---------------------------------------------------------------

    def get_price_by_type(self, price_type: PriceType) -> Decimal:
        raise NotImplementedError

    cdef object c_get_mid_price(self):
        raise NotImplementedError

    @property
    def ready(self) -> bool:
        raise NotImplementedError

    @property
    def market(self) -> ExchangeBase:
        raise NotImplementedError



from decimal import Decimal
from typing import Literal, Optional

from hummingbot.strategy_v2.executors.data_types import ConnectorPair, ExecutorConfigBase


class ArbitrageExecutorConfig(ExecutorConfigBase):
    type: Literal["arbitrage_executor"] = "arbitrage_executor"
    buying_market: ConnectorPair
    selling_market: ConnectorPair
    order_amount: Decimal
    min_profitability: Decimal
    gas_conversion_price: Optional[Decimal] = None



import asyncio
from decimal import Decimal
from typing import Any, Dict, List, Optional, Tuple

from bidict import bidict

from hummingbot.connector.constants import s_decimal_NaN
from hummingbot.connector.derivative.gate_io_perpetual import (
    gate_io_perpetual_constants as CONSTANTS,
    gate_io_perpetual_web_utils as web_utils,
)
from hummingbot.connector.derivative.gate_io_perpetual.gate_io_perpetual_api_order_book_data_source import (
    GateIoPerpetualAPIOrderBookDataSource,
)
from hummingbot.connector.derivative.gate_io_perpetual.gate_io_perpetual_auth import GateIoPerpetualAuth
from hummingbot.connector.derivative.gate_io_perpetual.gate_io_perpetual_user_stream_data_source import (
    GateIoPerpetualAPIUserStreamDataSource,
)
from hummingbot.connector.derivative.position import Position
from hummingbot.connector.perpetual_derivative_py_base import PerpetualDerivativePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair
from hummingbot.core.clock import Clock
from hummingbot.core.data_type.common import OrderType, PositionMode, PositionSide, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.core.web_assistant.connections.data_types import RESTMethod
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class GateIoPerpetualDerivative(PerpetualDerivativePyBase):
    """
    GateIoPerpetualExchange connects with Gate.io Derivative and provides order book pricing, user account tracking and
    trading functionality.
    """
    DEFAULT_DOMAIN = ""

    # Using 120 seconds here as Gate.io websocket is quiet
    TICK_INTERVAL_LIMIT = 120.0

    web_utils = web_utils

    def __init__(self,
                 gate_io_perpetual_api_key: str,
                 gate_io_perpetual_secret_key: str,
                 gate_io_perpetual_user_id: str,
                 balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
                 rate_limits_share_pct: Decimal = Decimal("100"),
                 trading_pairs: Optional[List[str]] = None,
                 trading_required: bool = True,
                 domain: str = DEFAULT_DOMAIN):
        """
        :param gate_io_perpetual_api_key: The API key to connect to private Gate.io APIs.
        :param gate_io_perpetual_secret_key: The API secret.
        :param trading_pairs: The market trading pairs which to track order book data.
        :param trading_required: Whether actual trading is needed.
        """
        self._gate_io_perpetual_api_key = gate_io_perpetual_api_key
        self._gate_io_perpetual_secret_key = gate_io_perpetual_secret_key
        self._gate_io_perpetual_user_id = gate_io_perpetual_user_id
        self._domain = domain
        self._position_mode = None
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs

        super().__init__(balance_asset_limit, rate_limits_share_pct)

        self._real_time_balance_update = False

    @property
    def authenticator(self):
        return GateIoPerpetualAuth(
            api_key=self._gate_io_perpetual_api_key,
            secret_key=self._gate_io_perpetual_secret_key)

    @property
    def name(self) -> str:
        return "gate_io_perpetual"

    @property
    def rate_limits_rules(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self):
        return self._domain

    @property
    def client_order_id_max_length(self):
        return CONSTANTS.MAX_ID_LEN

    @property
    def client_order_id_prefix(self):
        return CONSTANTS.HBOT_ORDER_ID

    @property
    def trading_rules_request_path(self):
        return CONSTANTS.EXCHANGE_INFO_URL

    @property
    def trading_pairs_request_path(self):
        return CONSTANTS.EXCHANGE_INFO_URL

    @property
    def check_network_request_path(self):
        return CONSTANTS.NETWORK_CHECK_PATH_URL

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return True

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    @property
    def funding_fee_poll_interval(self) -> int:
        return 120

    def _format_amount_to_size(self, trading_pair, amount: Decimal) -> Decimal:
        trading_rule = self._trading_rules[trading_pair]
        quanto_multiplier = Decimal(trading_rule.min_base_amount_increment)
        size = amount / quanto_multiplier
        return size

    def _format_size_to_amount(self, trading_pair, size: Decimal) -> Decimal:
        trading_rule = self._trading_rules[trading_pair]
        quanto_multiplier = Decimal(trading_rule.min_base_amount_increment)
        amount = size * quanto_multiplier
        return amount

    def supported_order_types(self) -> List[OrderType]:
        """
        :return a list of OrderType supported by this connector.
        Note that Market order type is no longer required and will not be used.
        """
        return [OrderType.LIMIT, OrderType.MARKET, OrderType.LIMIT_MAKER]

    def supported_position_modes(self):
        """
        This method needs to be overridden to provide the accurate information depending on the exchange.
        """
        return [PositionMode.ONEWAY, PositionMode.HEDGE]

    def get_buy_collateral_token(self, trading_pair: str) -> str:
        trading_rule: TradingRule = self._trading_rules[trading_pair]
        return trading_rule.buy_order_collateral_token

    def get_sell_collateral_token(self, trading_pair: str) -> str:
        trading_rule: TradingRule = self._trading_rules[trading_pair]
        return trading_rule.sell_order_collateral_token

    def start(self, clock: Clock, timestamp: float):
        """
        This function is called automatically by the clock.
        """
        super().start(clock, timestamp)

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        # API documentation does not clarify the error message for timestamp related problems
        return False

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        # TODO: implement this method correctly for the connector
        # The default implementation was added when the functionality to detect not found orders was introduced in the
        # ExchangePyBase class. Also fix the unit test test_lost_order_removed_if_not_found_during_order_status_update
        # when replacing the dummy implementation
        return False

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        # TODO: implement this method correctly for the connector
        # The default implementation was added when the functionality to detect not found orders was introduced in the
        # ExchangePyBase class. Also fix the unit test test_cancel_order_not_found_in_the_exchange when replacing the
        # dummy implementation
        return False

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            auth=self._auth)

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return GateIoPerpetualAPIOrderBookDataSource(
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return GateIoPerpetualAPIUserStreamDataSource(
            auth=self._auth,
            user_id=self._gate_io_perpetual_user_id,
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    async def start_network(self):
        """
        Start all required tasks to update the status of the connector.
        """
        await self._update_trading_rules()
        await super().start_network()

    async def _format_trading_rules(self, raw_trading_pair_info) -> List[TradingRule]:
        """
        Converts json API response into a dictionary of trading rules.
        :param symbols_info: The json API response
        :return A dictionary of trading rules.
        Response Example:
        [
            {
                "name": "BTC_USDT",
                "type": "direct",
                "quanto_multiplier": "0.0001",
                "ref_discount_rate": "0",
                "order_price_deviate": "0.5",
                "maintenance_rate": "0.005",
                "mark_type": "index",
                "last_price": "38026",
                "mark_price": "37985.6",
                "index_price": "37954.92",
                "funding_rate_indicative": "0.000219",
                "mark_price_round": "0.01",
                "funding_offset": 0,
                "in_delisting": false,
                "risk_limit_base": "1000000",
                "interest_rate": "0.0003",
                "order_price_round": "0.1",
                "order_size_min": 1,
                "ref_rebate_rate": "0.2",
                "funding_interval": 28800,
                "risk_limit_step": "1000000",
                "leverage_min": "1",
                "leverage_max": "100",
                "risk_limit_max": "8000000",
                "maker_fee_rate": "-0.00025",
                "taker_fee_rate": "0.00075",
                "funding_rate": "0.002053",
                "order_size_max": 1000000,
                "funding_next_apply": 1610035200,
                "short_users": 977,
                "config_change_time": 1609899548,
                "trade_size": 28530850594,
                "position_size": 5223816,
                "long_users": 455,
                "funding_impact_value": "60000",
                "orders_limit": 50,
                "trade_id": 10851092,
                "orderbook_id": 2129638396
              }
        ]
        """
        result = {}

        for rule in raw_trading_pair_info:
            try:
                if not web_utils.is_exchange_information_valid(rule):
                    continue

                trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=rule["name"])

                min_amount_inc = Decimal(f"{rule['quanto_multiplier']}")
                min_price_inc = Decimal(f"{rule['order_price_round']}")
                min_amount = min_amount_inc
                min_notional = Decimal(str(1))
                result[trading_pair] = TradingRule(trading_pair,
                                                   min_order_size=min_amount,
                                                   min_price_increment=min_price_inc,
                                                   min_base_amount_increment=min_amount_inc,
                                                   min_notional_size=min_notional,
                                                   min_order_value=min_notional,
                                                   )
            except Exception:
                self.logger().error(f"Error parsing the trading pair rule {rule}. Skipping.", exc_info=True)
        return list(result.values())

    async def _place_order(self,
                           order_id: str,
                           trading_pair: str,
                           amount: Decimal,
                           trade_type: TradeType,
                           order_type: OrderType,
                           price: Decimal,
                           **kwargs) -> Tuple[str, float]:
        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        size = self._format_amount_to_size(trading_pair, amount)
        data = {
            "text": order_id,
            "contract": symbol,
            "size": float(-size) if trade_type.name.lower() == 'sell' else float(size),
        }
        if order_type.is_limit_type():
            data.update({
                "price": f"{price:f}",
                "tif": "gtc",
            })
            if order_type is OrderType.LIMIT_MAKER:
                data.update({"tif": "poc"})
        else:
            data.update({
                "price": "0",
                "tif": "ioc",
            })

        # RESTRequest does not support json, and if we pass a dict
        # the underlying aiohttp will encode it to params
        data = data
        endpoint = CONSTANTS.ORDER_CREATE_PATH_URL
        order_result = await self._api_post(
            path_url=endpoint,
            data=data,
            is_auth_required=True,
            limit_id=endpoint,
        )
        if order_result.get('finish_as') in {"cancelled", "expired", "failed", "ioc"}:
            raise IOError({"label": "ORDER_REJECTED", "message": "Order rejected."})
        exchange_order_id = str(order_result["id"])
        return exchange_order_id, self.current_timestamp

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        """
        This implementation-specific method is called by _cancel
        returns True if successful
        """
        canceled = False
        exchange_order_id = await tracked_order.get_exchange_order_id()
        resp = await self._api_delete(
            path_url=CONSTANTS.ORDER_DELETE_PATH_URL.format(id=exchange_order_id),
            is_auth_required=True,
            limit_id=CONSTANTS.ORDER_DELETE_LIMIT_ID,
        )
        canceled = resp.get("finish_as") == "cancelled"
        return canceled

    async def _update_balances(self):
        """
        Calls REST API to update total and available balances.
        """
        account_info = ""
        try:
            account_info = await self._api_get(
                path_url=CONSTANTS.USER_BALANCES_PATH_URL,
                is_auth_required=True,
                limit_id=CONSTANTS.USER_BALANCES_PATH_URL
            )
            self._process_balance_message(account_info)
        except Exception as e:
            self.logger().network(
                f"Unexpected error while fetching balance update - {str(e)}", exc_info=True,
                app_warning_msg=(f"Could not fetch balance update from {self.name_cap}"))
            raise e
        return account_info

    async def check_network(self) -> NetworkStatus:
        try:
            await self._api_request(
                method=RESTMethod.GET,
                path_url=CONSTANTS.NETWORK_CHECK_PATH_URL,
            )
        except asyncio.CancelledError:
            raise
        except Exception:
            return NetworkStatus.NOT_CONNECTED
        return NetworkStatus.CONNECTED

    def _process_balance_message(self, account):
        local_asset_names = set(self._account_balances.keys())
        remote_asset_names = set()

        asset_name = account["currency"]
        self._account_available_balances[asset_name] = Decimal(str(account["available"]))
        self._account_balances[asset_name] = Decimal(str(account["total"]))
        remote_asset_names.add(asset_name)
        asset_names_to_remove = local_asset_names.difference(remote_asset_names)
        for asset_name in asset_names_to_remove:
            del self._account_available_balances[asset_name]
            del self._account_balances[asset_name]

    def _process_balance_message_ws(self, balance_update):
        for account in balance_update:
            asset_name = "USDT"
            self._account_available_balances[asset_name] = Decimal(str(account["balance"])) - Decimal(
                str(account["change"]))
            self._account_balances[asset_name] = Decimal(str(account["balance"]))

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        trade_updates = []

        try:
            exchange_order_id = await order.get_exchange_order_id()
            trading_pair = await self.exchange_symbol_associated_to_pair(trading_pair=order.trading_pair)
            all_fills_response = await self._api_get(
                path_url=CONSTANTS.MY_TRADES_PATH_URL,
                params={
                    "contract": trading_pair,
                    "order": exchange_order_id
                },
                is_auth_required=True,
                limit_id=CONSTANTS.MY_TRADES_PATH_URL)

            for trade_fill in all_fills_response:
                trade_update = self._create_trade_update_with_order_fill_data(
                    order_fill=trade_fill,
                    order=order)
                trade_updates.append(trade_update)

        except asyncio.TimeoutError:
            raise IOError(f"Skipped order update with order fills for {order.client_order_id} "
                          "- waiting for exchange order id.")

        return trade_updates

    def _create_trade_update_with_order_fill_data(
            self,
            order_fill: Dict[str, Any],
            order: InFlightOrder):
        fee_asset = order.quote_asset
        # no "position_action" in return, should use AddedToCostTradeFee, same as new_spot_fee
        fee = TradeFeeBase.new_spot_fee(
            fee_schema=self.trade_fee_schema(),
            trade_type=order.trade_type,
            percent_token=fee_asset,
            flat_fees=[TokenAmount(
                amount=Decimal(order_fill["fee"]),
                token=fee_asset
            )]
        )

        trade_update = TradeUpdate(
            trade_id=str(order_fill["id"]),
            client_order_id=order.client_order_id,
            exchange_order_id=order.exchange_order_id,
            trading_pair=order.trading_pair,
            fee=fee,
            fill_base_amount=abs(self._format_size_to_amount(order.trading_pair, (Decimal(str(order_fill["size"]))))),
            fill_quote_amount=abs(
                self._format_size_to_amount(order.trading_pair, (Decimal(str(order_fill["size"])))) * Decimal(
                    order_fill["price"])),
            fill_price=Decimal(order_fill["price"]),
            fill_timestamp=order_fill["create_time"],
        )
        return trade_update

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        try:
            exchange_order_id = await tracked_order.get_exchange_order_id()
            updated_order_data = await self._api_get(
                path_url=CONSTANTS.ORDER_STATUS_PATH_URL.format(id=exchange_order_id),
                is_auth_required=True,
                limit_id=CONSTANTS.ORDER_STATUS_LIMIT_ID)

            order_update = self._create_order_update_with_order_status_data(
                order_status=updated_order_data,
                order=tracked_order)
        except asyncio.TimeoutError:
            raise IOError(f"Skipped order status update for {tracked_order.client_order_id}"
                          f" - waiting for exchange order id.")

        return order_update

    def _create_order_update_with_order_status_data(self, order_status: Dict[str, Any], order: InFlightOrder):
        client_order_id = str(order_status.get("text", ""))
        state = self._normalise_order_message_state(order_status, order) or order.current_state

        order_update = OrderUpdate(
            trading_pair=order.trading_pair,
            update_timestamp=int(order_status["create_time"]),
            new_state=state,
            client_order_id=client_order_id,
            exchange_order_id=str(order_status["id"]),
        )
        return order_update

    def _normalise_order_message_state(self, order_msg: Dict[str, Any], tracked_order):
        state = None
        # we do not handle:
        #   "failed" because it is handled by create order
        #   "put" as the exchange order id is returned in the create order response
        #   "open" for same reason

        # same field for both WS and REST
        amount_left = Decimal(str(order_msg.get("left")))
        status = order_msg.get("status")
        finish_as = order_msg.get("finish_as")
        size = Decimal(str(order_msg.get("size")))
        if status == "finished":
            if finish_as == 'filled':
                state = OrderState.FILLED
            else:
                state = OrderState.CANCELED
        else:
            if amount_left > 0 and amount_left != size:
                state = OrderState.PARTIALLY_FILLED
        return state

    # use bybitperpetual sample,not gateio sample

    def _get_fee(self,
                 base_currency: str,
                 quote_currency: str,
                 order_type: OrderType,
                 order_side: TradeType,
                 amount: Decimal,
                 price: Decimal = s_decimal_NaN,
                 is_maker: Optional[bool] = None) -> TradeFeeBase:
        is_maker = is_maker or False
        fee = build_trade_fee(
            self.name,
            is_maker,
            base_currency=base_currency,
            quote_currency=quote_currency,
            order_type=order_type,
            order_side=order_side,
            amount=amount,
            price=price,
        )
        return fee

    async def _update_trading_fees(self):
        """
        Update fees information from the exchange
        """
        pass

    async def _user_stream_event_listener(self):
        """
        Listens to messages from _user_stream_tracker.user_stream queue.
        Traders, Orders, and Balance updates from the WS.
        """
        user_channels = [
            CONSTANTS.USER_TRADES_ENDPOINT_NAME,
            CONSTANTS.USER_ORDERS_ENDPOINT_NAME,
            CONSTANTS.USER_BALANCE_ENDPOINT_NAME,
            CONSTANTS.USER_POSITIONS_ENDPOINT_NAME,
        ]
        async for event_message in self._iter_user_event_queue():
            try:
                if isinstance(event_message, dict):
                    channel: str = event_message.get("channel", None)
                    results: List[Dict[str, Any]] = event_message.get("result", None)
                elif event_message is asyncio.CancelledError:
                    raise asyncio.CancelledError
                else:
                    raise Exception(event_message)
                if channel not in user_channels:
                    self.logger().error(
                        f"Unexpected message in user stream: {event_message}.", exc_info=True)
                    continue

                if channel == CONSTANTS.USER_TRADES_ENDPOINT_NAME:
                    for trade_msg in results:
                        self._process_trade_message(trade_msg)
                elif channel == CONSTANTS.USER_ORDERS_ENDPOINT_NAME:
                    for order_msg in results:
                        self._process_order_message(order_msg)
                elif channel == CONSTANTS.USER_POSITIONS_ENDPOINT_NAME:
                    for position_msg in results:
                        await self._process_account_position_message(position_msg)
                elif channel == CONSTANTS.USER_BALANCE_ENDPOINT_NAME:
                    self._process_balance_message_ws(results)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error(
                    "Unexpected error in user stream listener loop.", exc_info=True)
                await self._sleep(5.0)

    def _process_trade_message(self, trade: Dict[str, Any], client_order_id: Optional[str] = None):
        """
        Updates in-flight order and trigger order filled event for trade message received. Triggers order completed
        event if the total executed amount equals to the specified order amount.
        Example Trade:
        https://www.gate.io/docs/apiv4/en/#retrieve-market-trades
        """
        client_order_id = client_order_id or str(trade.get("text", ""))
        tracked_order = self._order_tracker.all_fillable_orders.get(client_order_id)

        if tracked_order is None:
            self.logger().debug(f"Ignoring trade message with id {client_order_id}: not in in_flight_orders.")
        else:
            trade_update = self._create_trade_update_with_order_fill_data(
                order_fill=trade,
                order=tracked_order)
            self._order_tracker.process_trade_update(trade_update)

    async def _process_account_position_message(self, position_msg: Dict[str, Any]):
        """
        Updates position
        :param position_msg: The position event message payload
        """
        ex_trading_pair = position_msg["contract"]
        trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=ex_trading_pair)
        amount = Decimal(str(position_msg["size"]))
        trading_rule = self._trading_rules[trading_pair]
        amount_precision = Decimal(trading_rule.min_base_amount_increment)
        position_side = PositionSide.LONG if Decimal(position_msg.get("size")) > 0 else PositionSide.SHORT
        pos_key = self._perpetual_trading.position_key(trading_pair, position_side)
        entry_price = Decimal(str(position_msg["entry_price"]))
        position = self._perpetual_trading.get_position(trading_pair, position_side)
        if position is not None:
            if amount == Decimal("0"):
                self._perpetual_trading.remove_position(pos_key)
            else:
                position.update_position(position_side=position_side,
                                         unrealized_pnl=None,
                                         entry_price=entry_price,
                                         amount=amount * amount_precision)
        else:
            await self._update_positions()

    def _process_order_message(self, order_msg: Dict[str, Any]):
        """
        Updates in-flight order and triggers cancelation or failure event if needed.

        :param order_msg: The order response from either REST or web socket API (they are of the same format)

        Example Order:
        https://www.gate.io/docs/apiv4/en/#list-orders
        """
        client_order_id = str(order_msg.get("text", ""))
        tracked_order = self._order_tracker.all_updatable_orders.get(client_order_id)
        if not tracked_order:
            self.logger().debug(f"Ignoring order message with id {client_order_id}: not in in_flight_orders.")
            return

        order_update = self._create_order_update_with_order_status_data(order_status=order_msg, order=tracked_order)
        self._order_tracker.process_order_update(order_update=order_update)

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: Dict[str, Any]):
        mapping = bidict()
        for symbol_data in filter(web_utils.is_exchange_information_valid, exchange_info):
            exchange_symbol = symbol_data["name"]
            base = symbol_data["name"].split('_')[0]
            quote = symbol_data["name"].split('_')[1]
            trading_pair = combine_to_hb_trading_pair(base, quote)
            if trading_pair in mapping.inverse:
                self._resolve_trading_pair_symbols_duplicate(mapping, exchange_symbol, base, quote)
            else:
                mapping[exchange_symbol] = trading_pair
        self._set_trading_pair_symbol_map(mapping)

    def _resolve_trading_pair_symbols_duplicate(self, mapping: bidict, new_exchange_symbol: str, base: str, quote: str):
        """Resolves name conflicts provoked by futures contracts.

        If the expected BASEQUOTE combination matches one of the exchange symbols, it is the one taken, otherwise,
        the trading pair is removed from the map and an error is logged.
        """
        expected_exchange_symbol = f"{base}_{quote}"
        trading_pair = combine_to_hb_trading_pair(base, quote)
        current_exchange_symbol = mapping.inverse[trading_pair]
        if current_exchange_symbol == expected_exchange_symbol:
            pass
        elif new_exchange_symbol == expected_exchange_symbol:
            mapping.pop(current_exchange_symbol)
            mapping[new_exchange_symbol] = trading_pair
        else:
            self.logger().error(
                f"Could not resolve the exchange symbols {new_exchange_symbol} and {current_exchange_symbol}")
            mapping.pop(current_exchange_symbol)

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        params = {
            "contract": await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        }

        resp_json = await self._api_request(
            method=RESTMethod.GET,
            path_url=CONSTANTS.TICKER_PATH_URL,
            params=params
        )

        return float(resp_json[0]["last"])

    async def _update_positions(self):
        """
        Retrieves all positions using the REST API.
        """

        positions = await self._api_get(
            path_url=CONSTANTS.POSITION_INFORMATION_URL,
            is_auth_required=True,
            limit_id=CONSTANTS.POSITION_INFORMATION_URL
        )

        for position in positions:
            ex_trading_pair = position.get("contract")
            hb_trading_pair = await self.trading_pair_associated_to_exchange_symbol(ex_trading_pair)

            amount = Decimal(position.get("size"))
            ex_mode = position.get("mode")
            if ex_mode == 'single':
                mode = PositionMode.ONEWAY
                position_side = PositionSide.LONG if Decimal(position.get("size")) > 0 else PositionSide.SHORT
            else:
                mode = PositionMode.HEDGE
                position_side = PositionSide.LONG if ex_mode == "dual_long" else PositionSide.SHORT
            pos_key = self._perpetual_trading.position_key(hb_trading_pair, position_side, mode)

            if amount != 0:
                trading_rule = self._trading_rules[hb_trading_pair]
                amount_precision = Decimal(trading_rule.min_base_amount_increment)

                unrealized_pnl = Decimal(position.get("unrealised_pnl"))
                entry_price = Decimal(position.get("entry_price"))
                leverage = Decimal(position.get("leverage"))
                position = Position(
                    trading_pair=hb_trading_pair,
                    position_side=position_side,
                    unrealized_pnl=unrealized_pnl,
                    entry_price=entry_price,
                    amount=amount * amount_precision,
                    leverage=leverage,
                )
                self._perpetual_trading.set_position(pos_key, position)

            else:
                self._perpetual_trading.remove_position(pos_key)

    async def _get_position_mode(self) -> Optional[PositionMode]:
        if self._position_mode is None:
            response = await self._api_get(
                path_url=CONSTANTS.POSITION_INFORMATION_URL,
                is_auth_required=True,
                limit_id=CONSTANTS.POSITION_INFORMATION_URL
            )
            self._position_mode = PositionMode.ONEWAY if response[0]["mode"] == 'single' else PositionMode.HEDGE
        return self._position_mode

    async def _execute_set_position_mode_for_pairs(
            # To-do: ensure there's no active order or contract before changing position mode
            self, mode: PositionMode, trading_pairs: List[str]
    ) -> Tuple[bool, List[str], str]:
        successful_pairs = []
        success = True
        msg = ""

        for trading_pair in trading_pairs:
            initial_mode = await self._get_position_mode()
            if mode != initial_mode:
                success, msg = await self._trading_pair_position_mode_set(mode, trading_pair)
                self._position_mode = mode
            if success:
                successful_pairs.append(trading_pair)
            else:
                self.logger().network(f"Error switching {trading_pair} mode to {mode}: {msg}")
                break

        return success, successful_pairs, msg

    async def _trading_pair_position_mode_set(self, mode: PositionMode, trading_pair: str) -> Tuple[bool, str]:
        msg = ""
        success = True

        dual_mode = 'true' if mode is PositionMode.HEDGE else 'false'

        data = {"dual_mode": dual_mode}

        response = await self._api_post(
            path_url=CONSTANTS.SET_POSITION_MODE_URL,
            params=data,
            is_auth_required=True,
            limit_id=CONSTANTS.SET_POSITION_MODE_URL,
        )
        if 'detail' in response:
            success = False
            msg = response['detail']
        return success, msg

    async def _set_trading_pair_leverage(self, trading_pair: str, leverage: int) -> Tuple[bool, str]:
        success = True
        msg = ""
        exchange_symbol = await self.exchange_symbol_associated_to_pair(trading_pair)
        if self.position_mode is PositionMode.ONEWAY:
            endpoint = CONSTANTS.ONEWAY_SET_LEVERAGE_PATH_URL.format(contract=exchange_symbol)
        else:
            endpoint = CONSTANTS.HEDGE_SET_LEVERAGE_PATH_URL.format(contract=exchange_symbol)
        data = {
            "leverage": leverage,
        }
        resp = await self._api_post(
            path_url=endpoint,
            params=data,
            is_auth_required=True,
            limit_id=CONSTANTS.ONEWAY_SET_LEVERAGE_PATH_URL,
        )
        if isinstance(resp, dict):
            return_leverage = resp['leverage']
        else:
            return_leverage = resp[0]['leverage']
        if int(return_leverage) != leverage:
            success = False
            msg = "leverage is diff"
        return success, msg

    async def _fetch_last_fee_payment(self, trading_pair: str) -> Tuple[int, Decimal, Decimal]:
        pass

    async def _update_funding_payment(self, trading_pair: str, fire_event_on_new: bool) -> bool:
        return True



import logging
import time
from typing import List, Optional

from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.data_feed.candles_feed.candles_base import CandlesBase
from hummingbot.data_feed.candles_feed.kraken_spot_candles import constants as CONSTANTS
from hummingbot.logger import HummingbotLogger


class KrakenSpotCandles(CandlesBase):
    _logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self, trading_pair: str, interval: str = "1m", max_records: int = 720):
        if max_records > 720:
            raise Exception("Kraken only supports a maximum of 720 records.")
        super().__init__(trading_pair, interval, max_records)

    @property
    def name(self):
        return f"kraken_{self._trading_pair}"

    @property
    def rest_url(self):
        return CONSTANTS.REST_URL

    @property
    def wss_url(self):
        return CONSTANTS.WSS_URL

    @property
    def health_check_url(self):
        return self.rest_url + CONSTANTS.HEALTH_CHECK_ENDPOINT

    @property
    def candles_url(self):
        return self.rest_url + CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_endpoint(self):
        return CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_max_result_per_rest_request(self):
        return CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST

    @property
    def rate_limits(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def intervals(self):
        return CONSTANTS.INTERVALS

    async def check_network(self) -> NetworkStatus:
        rest_assistant = await self._api_factory.get_rest_assistant()
        await rest_assistant.execute_request(url=self.health_check_url,
                                             throttler_limit_id=CONSTANTS.HEALTH_CHECK_ENDPOINT)
        return NetworkStatus.CONNECTED

    @staticmethod
    def convert_to_exchange_symbol(symbol: str) -> str:
        inverted_kraken_to_hb_map = {v: k for k, v in CONSTANTS.KRAKEN_TO_HB_MAP.items()}
        return inverted_kraken_to_hb_map.get(symbol, symbol)

    def get_exchange_trading_pair(self, hb_trading_pair: str, delimiter: str = "") -> str:
        """
        Note: The result of this method can safely be used to submit/make queries.
        Result shouldn't be used to parse responses as Kraken add special formating to most pairs.
        """
        if "-" in hb_trading_pair:
            base, quote = hb_trading_pair.split("-")
        elif "/" in hb_trading_pair:
            base, quote = hb_trading_pair.split("/")
        else:
            return hb_trading_pair
        base = self.convert_to_exchange_symbol(base)
        quote = self.convert_to_exchange_symbol(quote)

        exchange_trading_pair = f"{base}{delimiter}{quote}"
        return exchange_trading_pair

    @property
    def _is_first_candle_not_included_in_rest_request(self):
        return True

    @property
    def _is_last_candle_not_included_in_rest_request(self):
        return False

    def _get_rest_candles_params(self,
                                 start_time: Optional[int] = None,
                                 end_time: Optional[int] = None,
                                 limit: Optional[int] = CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST) -> dict:
        """
        For API documentation, please refer to:
        https://docs.kraken.com/rest/#tag/Spot-Market-Data/operation/getOHLCData

        This endpoint allows you to return up to 3600 candles ago.
        """
        candles_ago = (int(time.time()) - start_time) // self.interval_in_seconds
        if candles_ago > CONSTANTS.MAX_CANDLES_AGO:
            raise ValueError("Kraken REST API does not support fetching more than 720 candles ago.")
        return {"pair": self._ex_trading_pair, "interval": CONSTANTS.INTERVALS[self.interval],
                "since": start_time}

    def _parse_rest_candles(self, data: dict, end_time: Optional[int] = None) -> List[List[float]]:
        data: List = next(iter(data["result"].values()))
        new_hb_candles = []
        for i in data:
            timestamp = self.ensure_timestamp_in_seconds(float(i[0]))
            open = i[1]
            high = i[2]
            low = i[3]
            close = i[4]
            volume = i[6]
            quote_asset_volume = float(volume) * float(i[5])
            n_trades = 0
            taker_buy_base_volume = 0
            taker_buy_quote_volume = 0
            new_hb_candles.append([timestamp, open, high, low, close, volume,
                                   quote_asset_volume, n_trades, taker_buy_base_volume,
                                   taker_buy_quote_volume])
        return [candle for candle in new_hb_candles]

    def ws_subscription_payload(self):
        return {
            "event": "subscribe",
            "pair": [self.get_exchange_trading_pair(self._trading_pair, '/')],
            "subscription": {"name": CONSTANTS.WS_CANDLES_ENDPOINT,
                             "interval": int(CONSTANTS.INTERVALS[self.interval])}
        }

    def _parse_websocket_message(self, data: dict):
        candles_row_dict = {}
        if not (type(data) is dict and "event" in data.keys() and
                data["event"] in ["heartbeat", "systemStatus", "subscriptionStatus"]):
            if data[-2][:4] == "ohlc":
                candles_row_dict["timestamp"] = self.ensure_timestamp_in_seconds(data[1][1]) - self.interval_in_seconds
                candles_row_dict["open"] = data[1][2]
                candles_row_dict["high"] = data[1][3]
                candles_row_dict["low"] = data[1][4]
                candles_row_dict["close"] = data[1][5]
                candles_row_dict["volume"] = data[1][7]
                candles_row_dict["quote_asset_volume"] = float(data[1][7]) * float(data[1][6])
                candles_row_dict["n_trades"] = 0
                candles_row_dict["taker_buy_base_volume"] = 0
                candles_row_dict["taker_buy_quote_volume"] = 0
                return candles_row_dict



from decimal import Decimal
from math import exp, log
from typing import List


class Distributions:
    """
    A utility class containing methods to generate various types of numeric distributions.
    """

    @classmethod
    def linear(cls, n_levels: int, start: float = 0.0, end: float = 1.0) -> List[Decimal]:
        """
        Generate a linear sequence of spreads.

        Parameters:
        - n_levels: The number of spread levels to be generated.
        - start: The starting value of the sequence.
        - end: The ending value of the sequence.

        Returns:
        List[Decimal]: A list containing the generated linear sequence.
        """
        if n_levels == 1:
            return [Decimal(start)]

        return [Decimal(start) + (Decimal(end) - Decimal(start)) * Decimal(i) / (Decimal(n_levels) - 1) for i in range(n_levels)]

    @classmethod
    def fibonacci(cls, n_levels: int, start: float = 0.01) -> List[Decimal]:
        """
        Generate a Fibonacci sequence of spreads represented as percentages.

        The Fibonacci sequence is a series of numbers in which each number (Fibonacci number)
        is the sum of the two preceding ones. In this implementation, the sequence starts with
        the provided initial_value (represented as a percentage) and the value derived by adding
        the initial_value to itself as the first two terms. Each subsequent term is derived by
        adding the last two terms of the sequence.

        Parameters:
        - n_levels (int): The number of spread levels to be generated.
        - initial_value (float, default=0.01): The value from which the Fibonacci sequence will start,
          represented as a percentage. Default is 1%.

        Returns:
        List[Decimal]: A list containing the generated Fibonacci sequence of spreads, represented as percentages.

        Example:
        If initial_value=0.01 and n_levels=5, the sequence would represent: [1%, 2%, 3%, 5%, 8%]
        """

        if n_levels == 1:
            return [Decimal(start)]

        fib_sequence = [Decimal(start), Decimal(start) * 2]
        for i in range(2, n_levels):
            fib_sequence.append(fib_sequence[-1] + fib_sequence[-2])
        return fib_sequence[:n_levels]

    @classmethod
    def logarithmic(cls, n_levels: int, base: float = exp(1), scaling_factor: float = 1.0,
                    start: float = 0.4) -> List[Decimal]:
        """
        Generate a logarithmic sequence of spreads.

        Parameters:
        - n_levels: The number of spread levels to be generated.
        - base: The base value for the logarithm. Default is Euler's number.
        - scaling_factor: The factor to scale the logarithmic value.
        - initial_value: Initial value for translation.

        Returns:
        List[Decimal]: A list containing the generated logarithmic sequence.
        """
        translation = Decimal(start) - Decimal(scaling_factor) * Decimal(log(2, base))
        return [Decimal(scaling_factor) * Decimal(log(i + 2, base)) + translation for i in range(n_levels)]

    @classmethod
    def arithmetic(cls, n_levels: int, start: float, step: float) -> List[Decimal]:
        """
        Generate an arithmetic sequence of spreads.

        Parameters:
        - n_levels: The number of spread levels to be generated.
        - start: The starting value of the sequence.
        - increment: The constant value to be added in each iteration.

        Returns:
        List[Decimal]: A list containing the generated arithmetic sequence.
        """
        return [Decimal(start) + i * Decimal(step) for i in range(n_levels)]

    @classmethod
    def geometric(cls, n_levels: int, start: float, ratio: float) -> List[Decimal]:
        """
        Generate a geometric sequence of spreads.

        Parameters:
        - n_levels: The number of spread levels to be generated.
        - start: The starting value of the sequence.
        - ratio: The ratio to multiply the current value in each iteration. Should be greater than 1 for increasing sequence.

        Returns:
        List[Decimal]: A list containing the generated geometric sequence.
        """
        if ratio <= 1:
            raise ValueError(
                "Ratio for modified geometric distribution should be greater than 1 for increasing spreads.")

        return [Decimal(start) * Decimal(ratio) ** Decimal(i) for i in range(n_levels)]



from bidict import bidict

from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit

REST_URL = "https://api.pacifica.fi/api/v1"
WSS_URL = "wss://ws.pacifica.fi/ws"

HEALTH_CHECK_ENDPOINT = "/info"
CANDLES_ENDPOINT = "/kline"

WS_CANDLES_CHANNEL = "candle"

# Supported intervals based on Pacifica's WebSocket documentation
# 1m, 3m, 5m, 15m, 30m, 1h, 2h, 4h, 8h, 12h, 1d
INTERVALS = bidict({
    "1m": "1m",
    "3m": "3m",
    "5m": "5m",
    "15m": "15m",
    "30m": "30m",
    "1h": "1h",
    "2h": "2h",
    "4h": "4h",
    "8h": "8h",
    "12h": "12h",
    "1d": "1d",
})

MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST = 1000

# Rate limits for candles feed (public data - no authentication required)
# Using Unidentified IP tier: 125 credits/60s
# Credit costs for unidentified IP:
# - Standard request: 1 credit
# - Heavy GET requests: 3-12 credits
# Documentation: https://docs.pacifica.fi/api-documentation/api/rate-limits
PACIFICA_CANDLES_LIMIT_ID = "PACIFICA_CANDLES_LIMIT"
HEAVY_GET_REQUEST_COST = 12  # Conservative estimate for heavy GET (max for unidentified IP)

RATE_LIMITS = [
    RateLimit(limit_id=PACIFICA_CANDLES_LIMIT_ID, limit=125, time_interval=60),
    RateLimit(limit_id=HEALTH_CHECK_ENDPOINT, limit=125, time_interval=60,
              linked_limits=[LinkedLimitWeightPair(PACIFICA_CANDLES_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST)]),
    RateLimit(limit_id=CANDLES_ENDPOINT, limit=125, time_interval=60,
              linked_limits=[LinkedLimitWeightPair(PACIFICA_CANDLES_LIMIT_ID, weight=HEAVY_GET_REQUEST_COST)]),
]
